import itertools
import warnings
from multiprocessing import Pool

import PIL
import cv2
import numpy as np
from PIL import Image, Image as im
from matplotlib import pyplot as plt

from eidl.utils.iter_utils import reverse_tuple


def generate_image_binary_mask(image, channel_first=False):
    if channel_first:
        image = np.moveaxis(image, 0, -1)
    # Convert the RGB image to grayscale
    gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    _, binary_mask = cv2.threshold(gray_image, 254, 1, cv2.THRESH_BINARY_INV)
    return binary_mask

def z_normalize_image(image, mean, std):
    assert image.shape[-1] == 3, "Image should be in channel last format"
    image = image.astype(np.float32)
    image -= mean
    image /= std
    return image


def resize_image(image_name, image_size, image):
    image = load_oct_image(image, image_size)
    return image_name, {'image': image}


def load_oct_image(image_info, image_size):
    """

    @param image_info:
        if str, interpret as the image path,
        if dict, interpret as the image info dict, comes with the cropped image data
    """
    if isinstance(image_info, str):
        image = Image.open(image_info).convert('RGB')
    elif isinstance(image_info, np.ndarray):
        image = image_info
    else:
        raise ValueError(f"image info {image_info} is not a valid type")
    # image = image.crop((0, 0, 5360, 2656))
    # image = image.crop((0, 0, 5120, 2640))
    image = im.fromarray(image).resize(image_size, resample=PIL.Image.LANCZOS)
    image = np.array(image).astype(np.float32)
    return image

def crop_image(image, patch_size):
    """
    crop the image into patches of size patch_size
    @param image:
    @param patch_size:
    @return:
    """
    image_size = image.shape[:2]
    # crop the image into patches of size patch_size
    n_patch_rows = image_size[0] // patch_size[0]
    n_patch_cols = image_size[1] // patch_size[1]
    return image[:n_patch_rows * patch_size[0], :n_patch_cols * patch_size[1]]  # crop from the bottom right corner


def pad_image(image, max_n_patches, patch_size):

    image_size = image.shape[:2]

    # image shape must be divisible by patch size before padding
    assert image_size[0] % patch_size[0] == 0 and image_size[1] % patch_size[1] == 0, f"image shape {image_size} is not divisible by patch size {patch_size}"
    n_patches = (image_size[0] // patch_size[0], image_size[1] // patch_size[1])

    # pad the image to the max size
    pad_size = ((max_n_patches[0] - n_patches[0]) * patch_size[0], (max_n_patches[1] - n_patches[1]) * patch_size[1])
    image_padded = np.pad(image, ((0, pad_size[0]), (0, pad_size[1]), (0, 0)), mode='constant', constant_values=0)

    patch_mask = np.ones((max_n_patches[0], max_n_patches[1]), dtype=bool)
    patch_mask[n_patches[0]:, :] = False
    patch_mask[:, n_patches[1]:] = False

    return image_padded, patch_mask


def preprocess_subimages(cropped_image_data, patch_size=(32, 32), white_patch_mask_threshold=.95, *args, **kwargs):
    """
    sub image pad to max size
        'En-face_52.0micrometer_Slab_(Retina_View)':
        'Circumpapillary_RNFL':
        'RNFL_Thickness_(Retina_View)':
        'GCL_Thickness_(Retina_View)':
        'RNFL_Probability_and_VF_Test_points(Field_View)':
        'GCL+_Probability_and_VF_Test_points':

    first crop the image to the closest size divisible by the patch size, pad each sub image to the same size, so that we can batchify them

    Parameters
    ----------
    cropped_image_data
    patch_size: tuple of int, width and height of the patch
    white_patch_mask_threshold: a float between 0 and 1, if more than this percentage of the patch is white, then the patch is masked out.
                                if 0, then all patches are masked out, if 1, then a patch has to be all white to be masked out

    Returns
    -------

    """
    image_names = list(cropped_image_data.keys())
    sub_image_names = list(cropped_image_data[image_names[0]]['sub_images'].keys())

    counter = 0
    for i, s_image_name in enumerate(sub_image_names):
        sub_images = {image_name: (image_data['sub_images'][s_image_name]['sub_image'], image_data['sub_images'][s_image_name]['position']) for image_name, image_data in cropped_image_data.items()}
        max_size = max([s_image.shape[:2] for (s_image, _) in sub_images.values()])
        max_size = (max_size[0] // patch_size[0] * patch_size[0], max_size[1] // patch_size[1] * patch_size[1])
        max_n_patches = (max_size[0] // patch_size[0], max_size[1] // patch_size[1])

        print(f"resizing sub-images {s_image_name}, {i + 1}/{len(sub_image_names)}, they will be cropped&padded to {max_size}, with {max_n_patches} patches ({patch_size=})")
        # find the max patchifiable size, round down

        for image_name, (s_image, position) in sub_images.items():
            temp = crop_image(s_image, patch_size)

            cropped_image_data[image_name]['sub_images'][s_image_name]['sub_image_cropped_padded'], patch_mask = pad_image(temp, max_n_patches, patch_size)
            cropped_image_data[image_name]['sub_images'][s_image_name]['position'] = position

            white_mask = generate_image_binary_mask(cropped_image_data[image_name]['sub_images'][s_image_name]['sub_image_cropped_padded'], channel_first=False)
            white_mask_patches = white_mask.reshape(white_mask.shape[0] // patch_size[0], patch_size[0], white_mask.shape[1] // patch_size[1], patch_size[1])
            white_mask_patches = white_mask_patches.transpose(0, 2, 1, 3)
            white_mask_patches = white_mask_patches.reshape(-1, *patch_size)
            white_mask_patches = [(True if np.mean(patch) > white_patch_mask_threshold else False) for patch in white_mask_patches]
            white_mask_patches = np.reshape(white_mask_patches, patch_mask.shape)
            patch_mask = np.logical_and(patch_mask, white_mask_patches)
            # add white and black masks
            cropped_image_data[image_name]['sub_images'][s_image_name]['patch_mask'] = patch_mask

            # plt.imsave(f'C:/Users/apoca/Downloads/temp/{counter}_{s_image_name}_Aoriginal_subimage.png', s_image)
            # plt.imsave(f'C:/Users/apoca/Downloads/temp/{counter}_{s_image_name}_Bimage_cropped.png', temp)
            # plt.imsave(f'C:/Users/apoca/Downloads/temp/{counter}_{s_image_name}_Cimage_padded.png', cropped_image_data[image_name]['sub_images'][s_image_name]['sub_image_cropped_padded'])
            # plt.imsave(f'C:/Users/apoca/Downloads/temp/{counter}_{s_image_name}_Dpatch_mask.png', cropped_image_data[image_name]['sub_images'][s_image_name]['patch_mask'])

            counter += 1
    return cropped_image_data


def patchify(image, patch_size):
    # Unpack the dimensions of the patch size
    patch_height, patch_width = patch_size

    # Get the dimensions of the image
    img_height, img_width = image.shape[:2]

    # Check if the image is grayscale or color
    if len(image.shape) == 2:  # Grayscale image
        # Reshape the image into patches
        patches = image.reshape(img_height // patch_height, patch_height, img_width // patch_width, patch_width)
        # Transpose the axes to bring patches to the front
        patches = patches.transpose(0, 2, 1, 3)
        # Reshape to the final 2D array (num_patches, patch_height, patch_width)
        patches = patches.reshape(-1, patch_height, patch_width)
    else:  # Color image
        # Get the number of channels
        num_channels = image.shape[2]
        # Reshape the image into patches
        patches = image.reshape(img_height // patch_height, patch_height, img_width // patch_width, patch_width, num_channels)
        # Transpose the axes to bring patches to the front
        patches = patches.transpose(0, 2, 1, 3, 4)
        # Reshape to the final 3D array (num_patches, patch_height, patch_width, num_channels)
        patches = patches.reshape(-1, patch_height, patch_width, num_channels)

    return patches


def z_norm_subimages(name_label_images_dict, n_jobs=1, *args, **kwargs):
    image_names = list(name_label_images_dict.keys())
    sub_image_names = list(name_label_images_dict[image_names[0]]['sub_images'].keys())

    all_sub_images = [image_data['sub_images'][s_image_name]['sub_image'] for image_name, image_data in name_label_images_dict.items() for i, s_image_name in enumerate(sub_image_names)]

    mean_values = np.stack([np.mean(image, axis=(0, 1)) for image in all_sub_images])
    all_mean = np.mean(mean_values, axis=0)

    std_values = np.stack([np.std(image, axis=(0, 1)) for image in all_sub_images])
    all_std = np.sqrt(np.mean(np.square(std_values), axis=0))

    # now normalize the sub images
    # znorm_image_args = [[(s_image_data['sub_image_cropped_padded'], all_mean, all_std) for s_image_name, s_image_data in image_data['sub_images'].items()] for image_name, image_data in name_label_images_dict.items()]
    # znorm_image_args = list(itertools.chain.from_iterable(znorm_image_args))
    # with Pool(n_jobs) as p:
    #     images = p.starmap(z_normalize_image, znorm_image_args)

    for image_name, image_data in name_label_images_dict.items():
        for s_image_name, s_image_data in image_data['sub_images'].items():
            s_image_data['sub_image_cropped_padded_z_normed'] = z_normalize_image(s_image_data['sub_image_cropped_padded'], all_mean, all_std)
            # s_image_data['sub_image_cropped_padded_z_normed'] = (s_image_data['sub_image_cropped_padded'] - all_mean) / all_std

    return name_label_images_dict, all_mean, all_std


def get_heatmap(seq, grid_size, normalize=True):
    """
    get the heatmap from the fixations
    Parameters

    grid_size: tuple of ints
    patch_size:
    ----------
    seq

    Returns
    -------

    """
    heatmap = np.zeros(grid_size)
    grid_height, grid_width = grid_size
    for i in seq:
        heatmap[int(np.floor(i[1] * grid_height)), int(np.floor(i[0] * grid_width))] += 1
    assert (heatmap.sum() == len(seq))
    if normalize:
        heatmap = heatmap / heatmap.sum()
        assert abs(heatmap.sum() - 1) < 0.01, ValueError("no fixations sequence")
    return heatmap


def process_aoi(aoi_heatmap, image_size, has_subimage, grid_size, **kwargs):
    if has_subimage:
        return remap_subimage_aoi(aoi_heatmap, image_size=image_size, **kwargs)
    else:
        aoi_heatmap = aoi_heatmap.reshape(grid_size)
        aoi_heatmap = cv2.resize(aoi_heatmap, dsize=image_size, interpolation=cv2.INTER_LINEAR)

def remap_subimage_aoi(subimage_patch_aoi, subimage_masks, subimages, subimage_positions, image_size, patch_size, normalize_by_subimage=False, **kwargs):
    """


    Parameters
    ----------
    subimage_patch_aoi: ndarray, 1D array of size (n_patches,)
    subimage_masks
    subimage_positions
    image_size

    Returns
    -------

    """
    aoi_recovered = np.zeros(image_size)
    sub_image_aois = []
    subimage_patch_counter = 0
    for s_image, s_mask, s_pos in zip(subimages, subimage_masks, subimage_positions):  # s refers to a single subimage
        s_patch_size = np.prod(s_mask.shape)
        s_aoi = subimage_patch_aoi[subimage_patch_counter:(subimage_patch_counter + s_patch_size)].reshape(s_mask.shape)

        s_image_size_cropped_or_padded = s_aoi.shape[0] * patch_size[0], s_aoi.shape[1] * patch_size[1]  # the aoi is padded
        s_image_size = s_pos[2][1] - s_pos[0][1], s_pos[2][0] - s_pos[0][0]

        # s_aoi = cv2.resize(s_aoi, dsize=reverse_tuple(s_image_size), interpolation=cv2.INTER_LINEAR)
        s_aoi = cv2.resize(s_aoi, dsize=reverse_tuple(s_image_size_cropped_or_padded), interpolation=cv2.INTER_LINEAR)

        s_aoi = s_aoi[:s_image_size[0], :s_image_size[1]]  # if the image is padded, this also remove the attention from the padded area

        if normalize_by_subimage:
            if np.max(s_aoi) > 0:
                s_aoi = s_aoi / np.max(s_aoi)
            else:
                warnings.warn(f"subimage {s_image} has no attention: zero division encounter when normalizing the attention. Nothing will be done to the attention.")
        aoi_recovered[s_pos[0][1]:min(s_pos[2][1], s_pos[0][1] + s_image_size_cropped_or_padded[0]),  # the min is dealing with the cropped case
                      s_pos[0][0]:min(s_pos[2][0], s_pos[0][0] + s_image_size_cropped_or_padded[1])] += s_aoi
        sub_image_aois.append(s_aoi)
        subimage_patch_counter += s_patch_size
    return aoi_recovered, sub_image_aois

def remap_subimage_attention_rolls(rolls, subimage_masks, subsubimage_positions, original_image_size):
    print("remapping subimage attention rolls")





import os
import pickle
import tempfile
import urllib

import numpy as np
import timm
import torch
import gdown

from eidl.Models.ExpertAttentionViT import ViT_LSTM
from eidl.Models.ExpertAttentionViTSubImages import ViT_LSTM_subimage
from eidl.Models.ExpertTimmVisionTransformer import ExpertTimmVisionTransformer
from eidl.Models.ExpertTimmVisionTransformerSubimage import ExpertTimmVisionTransformerSubimage
from eidl.Models.ExtensionModel import ExtensionModelSubimage
from eidl.utils.image_utils import load_oct_image
from eidl.utils.iter_utils import reverse_tuple, chunker


def get_model(model_name, image_size, depth, device, *args, **kwargs):
    # if type(image_size[0]) == int:
    #     image_size = swap_tuple(image_size, 0, -1)
    # if isinstance(image_size[0], Iterable):
    #     image_size = [swap_tuple(x, 0, -1) for x in image_size]
    if model_name == 'base':
        # model = ViT_LSTM(image_size=reverse_tuple(image_size), patch_size=(32, 16), num_classes=2, embed_dim=128, depth=depth, heads=1,
        #                  mlp_dim=2048, weak_interaction=False).to(device)
        model = ViT_LSTM(image_size=image_size, num_patches=32, num_classes=2, embed_dim=128, depth=depth, heads=1,
                         mlp_dim=2048, weak_interaction=False).to(device)
    elif model_name == 'base_subimage':
        model = ViT_LSTM_subimage(image_size=image_size, num_classes=2, embed_dim=128, depth=depth, heads=1,
                         mlp_dim=2048, weak_interaction=False, *args, **kwargs).to(device)  # NOTE, only this option supporst variable patch size
    elif model_name == 'vit_small_patch32_224_in21k':  # assuming any other name is timm models
        model = timm.create_model(model_name, img_size=reverse_tuple(image_size), pretrained=True, num_classes=2)  # weights from 'https://storage.googleapis.com/vit_models/augreg/L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1.npz', official Google JAX implementation
        model = ExpertTimmVisionTransformer(model).to(device)
    elif model_name == 'vit_small_patch32_224_in21k_subimage':
        model = timm.create_model(model_name.strip('_subimage'),  pretrained=True, num_classes=2, dynamic_img_size=True)  # weights from 'https://storage.googleapis.com/vit_models/augreg/L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1.npz', official Google JAX implementation
        model = ExpertTimmVisionTransformerSubimage(model).to(device)
    elif model_name == 'InceptionV4_subimage':
        model = timm.create_model(model_name.strip('_subimage'),  pretrained=True, dynamic_img_size=True, features_only=True)  # weights from 'https://storage.googleapis.com/vit_models/augreg/L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1.npz', official Google JAX implementation
        model = ExtensionModelSubimage(model, num_classes=2).to(device)
    else:
        raise ValueError(f"model name {model_name} is not supported")
    return model, model.get_grid_size()


def swap_tuple(t, i, j):
    t = list(t)
    t[i], t[j] = t[j], t[i]
    return tuple(t)

def parse_model_parameter(model_config_string: str, parameter_name: str):
    assert parameter_name in model_config_string
    parameter_string = [x for x in model_config_string.split('_') if parameter_name in x][0]
    parameter_value = parameter_string.split('-')[1]
    if parameter_name == 'dist':
        return parameter_string.strip(f'{parameter_name}-')
    elif parameter_name in ['alpha', 'dist', 'depth', 'lr']:
        return float(parameter_string.strip(f'{parameter_name}-'))
    elif parameter_name == 'model':
        return model_config_string[:model_config_string.find('_alpha')].split('-')[1]
    else:
        return parameter_value


def get_trained_model(device, model_param):
    """
    to use the model returned by this function, user should use model_utils.load_image and pass the returns (image_mean, image_std, image_size)
    as arguments.
    Parameters
    ----------
    device

    Returns
    a tuple of four items
    model: the trained model
    model_param: str: can be 'num-patch-32_image-size-1024-512', or 'patch-size-50-25_image-size-1000-500'
    image_mean: means of the RGB channels of the data on which the model is trained
    image_std: stds of the
    image_size: the size of the image used by the model
    -------

    """
    model_name = 'base'
    depth = 1

    if model_param == 'num-patch-32_image-size-1024-512':
        image_size = 1024, 512
    elif model_param == 'patch-size-50-25_image-size-1000-500':
        image_size = 1000, 500
    else:
        raise ValueError(f"model_param {model_param} is not supported")

    github_file_url = "https://raw.githubusercontent.com/ApocalyVec/ExpertInformedDL/master/trained_model/0.0.1"
    model_url = f"{github_file_url}/trained_model/best_model-base_alpha-0.01_dist-cross-entropy_depth-1_lr-0.0001_statedict_{model_param}.pt"
    image_mstd_url = f"{github_file_url}/image_means_stds_{model_param}.p"
    compound_label_encoder_url = f"{github_file_url}/compound_label_encoder.p"

    temp_dir = tempfile.mkdtemp()
    model_file_path = os.path.join(temp_dir, "model_weights.pt")
    image_mstd_file_path = os.path.join(temp_dir, f"image_means_stds_{model_param}.p")
    compound_label_encoder_file_path = os.path.join(temp_dir, "compound_label_encoder.p")

    # Download the file using urlretrieve
    urllib.request.urlretrieve(model_url, model_file_path)
    urllib.request.urlretrieve(image_mstd_url, image_mstd_file_path)
    urllib.request.urlretrieve(compound_label_encoder_url, compound_label_encoder_file_path)

    print(f"File downloaded successfully and saved to {model_file_path}")
    model, grid_size = get_model(model_name, image_size=image_size, depth=depth, device=device)
    model.load_state_dict(torch.load(model_file_path))

    image_mean, image_std = pickle.load(open(image_mstd_file_path, 'rb'))

    compound_label_encoder = pickle.load(open(compound_label_encoder_file_path, 'rb'))
    return model, image_mean, image_std, image_size, compound_label_encoder

def get_subimage_model(*args, **kwargs):
    github_file_url = "https://raw.githubusercontent.com/ApocalyVec/ExpertInformedDL/master/trained_model/0.0.2"
    model_url = f"{github_file_url}/model.pt"
    temp_dir = tempfile.gettempdir()
    model_file_path = os.path.join(temp_dir, "subimage_model.pt")
    dataset_path = os.path.join(temp_dir, "oct_reports_info.p")

    if not os.path.exists(model_file_path):
        urllib.request.urlretrieve(model_url, model_file_path)
    model = torch.load(model_file_path)
    print("Model downloaded and loaded.")

    patch_size = model.patch_height, model.patch_width

    # get the dataset
    if not os.path.exists(dataset_path):
        print("Downloading the dataset...")

        gdown.download("https://drive.google.com/uc?id=1V2-jSmEKl-7xzvleYRoDZJAdAQP7be_G", output=dataset_path, quiet=False)
    from eidl.utils.SubimageHandler import SubimageHandler
    data = pickle.load(open(dataset_path, 'rb'))
    subimage_handler = SubimageHandler()
    subimage_handler.load_image_data(data, patch_size=patch_size, *args, **kwargs)
    subimage_handler.model = model

    return subimage_handler


def load_image_preprocess(image_path, image_size, image_mean, image_std):
    image = load_oct_image(image_path, image_size)
    image_normalized = (image - image_mean) / image_std
    # transpose to channel first
    image_normalized = image_normalized.transpose((2, 0, 1))
    return image_normalized, image

def get_best_model(models, results_dict):
    models = list(reversed(list(models)))
    best_model, best_model_results, best_model_config_string = None, None, None
    for model in models:  # get the best model each model architecture
        # model = 'vit_small_patch32_224_in21k'
        # model = 'base'
        best_model_val_acc = -np.inf
        best_model_config_string = None
        best_model_results = None
        for model_config_string, results in results_dict.items():
            this_val_acc = np.max(results['val_accs'])
            if parse_model_parameter(model_config_string, 'model') == model and this_val_acc > best_model_val_acc:
                best_model_val_acc = this_val_acc
                best_model_config_string = model_config_string
                best_model_results = results

        print(f"Best model for {model} has val acc of {best_model_val_acc} with parameters: {best_model_config_string}")
        best_model = best_model_results['model']
    return best_model, best_model_results, best_model_config_string

def parse_training_results(results_dir):
    results_dict = {}
    model_config_strings = [i.strip('log_').strip('.txt') for i in os.listdir(results_dir) if i.startswith('log')]
    # columns = ['model_name', 'train acc', 'train loss', 'validation acc', 'validation loss', 'test acc']
    # results_df = pd.DataFrame(columns=columns)

    for i, model_config_string in enumerate(model_config_strings):
        print(f"Processing [{i}] of {len(model_config_strings)} model configurations: {model_config_string}")
        model = torch.load(
            os.path.join(results_dir,
                         f'best_{model_config_string}.pt'))  # TODO should load the model with smallest loss??
        with open(os.path.join(results_dir, f'log_{model_config_string}.txt'), 'r') as file:
            lines = file.readlines()

        results = []
        for epoch_lines in chunker(lines, 3):  # iterate three lines at a time
            train_loss, train_acc = [np.nan if x == '' else float(x) for x in epoch_lines[1].strip("training: ").split(",")]
            val_loss, val_acc = [np.nan if x == '' else float(x) for x in epoch_lines[2].strip('validation: ').split(",")]
            results.append((train_acc, train_loss, val_acc, val_loss))
        results = np.array(results)
        # best_val_acc_epoch_index = np.argmax(results[:, 2])
        # test_acc = test_without_fixation(model, test_loader, device)  # TODO restore the test_acc after adding test method to extention
        # add viz pca of patch embeddings, attention rollout (gif and overlay), and position embeddings,
        # values = [model_config_string, *results[best_val_acc_epoch_index], test_acc]
        # results_df = pd.concat([results_df, pd.DataFrame(dict(zip(columns, values)))], ignore_index=True) # TODO fix the concat
        test_acc = None
        results_dict[model_config_string] = {'model_config_string': model_config_string, 'train_accs': results[:, 0], 'train_losses': results[:, 1], 'val_accs': results[:, 2], 'val_losses': results[:, 3], 'test_acc': test_acc, 'model': model}  # also save the model
    # results_df.to_csv(os.path.join(results_dir, "summary.csv"))
    return results_dict, model_config_strings



import torch
from torch import nn

from einops import rearrange, repeat
from einops.layers.torch import Rearrange
import torch.nn.utils.rnn as rnn_utils

from eidl.Models.ExpertAttentionViT import ViT_LSTM


class ViT_LSTM_subimage(nn.Module):
    def __init__(self, *args, **kwargs):
        """

        Parameters
        ----------
        image_size
        num_classes
        embed_dim
        depth
        heads
        mlp_dim
        pool
        channels
        dim_head
        dropout
        emb_dropout
        weak_interaction
        num_patches: int: number of patches in each dimension, using this parameter will override patch_size and will create the
        the same number of patches across height and width
        patch_size: tuple: tuple of two integers, in pixels (height, width)
        """
        super().__init__()
        self.ViT = ViT_LSTM(*args, **kwargs)
        self.depth = self.ViT.depth
        self.patch_height, self.patch_width = self.ViT.patch_height, self.ViT.patch_width

    def forward(self, img, collapse_attention_matrix=True, *args, **kwargs):
        '''

        Parameters
        ----------
        img
        collapse_attention_matrix
        args
        kwargs

        Returns
        -------

        '''

        # apply patch embedding to each subimage

        # flatten the mask for the attention layer
        subimage_xs = [self.ViT.to_patch_embedding(x) for x in img['subimages']]
        mask = [torch.flatten(m, 1, 2) for m in img['masks']]
        mask = torch.cat(mask, dim=1)
        # concatenate the subimage patches
        x = torch.cat(subimage_xs, dim=1)

        return self.ViT._encode(x, collapse_attention_matrix=collapse_attention_matrix, mask=mask, *args, **kwargs)


    def get_grid_size(self):
        return self.ViT.get_grid_size()




import os
import pickle
import tempfile

import torch

from eidl.utils.SubimageHandler import SubimageHandler
from eidl.utils.model_utils import get_best_model, parse_training_results, parse_model_parameter

patch_size=(32, 32)
data_path = 'C:/Dropbox/ExpertViT/Datasets/OCTData/oct_v2/oct_reports_info.p'
# results_dir = 'results-01_05_2024_10_56_13'
results_dir = '../temp/results-01_03_2024_22_31_25'
source_attention_path = r"../temp/source_attention\GCL Prob RLS_036_OS_TC.pickle"
figure_dir = '../temp/figures_example/RLS_036_OS_TC'

# figure_notes = 'square depth 1'
# figure_notes = 'static aggregated-self discard 0.1 '
figure_notes = 'test '
if __name__ == '__main__':
    # load sample human attention ###################
    human_attention = pickle.load(open(source_attention_path, 'rb'))

    # load image data ###########################################################
    # the image data must comply with the format specified in SubimageLoader

    # check if the subimage handler is in the temp directory
    if os.path.exists(os.path.join(tempfile.gettempdir(), 'subimage_handler.p')):
        subimage_handler = pickle.load(open(os.path.join(tempfile.gettempdir(), 'subimage_handler.p'), 'rb'))
    else:
        data = pickle.load(open(data_path, 'rb'))
        subimage_handler = SubimageHandler()
        subimage_handler.load_image_data(data, patch_size=patch_size)
        # save the subimage handler to temp directory
        pickle.dump(subimage_handler, open(os.path.join(tempfile.gettempdir(), 'subimage_handler.p'), 'wb'))

    # load model ###############################################################
    # find the best model in result directory
    results_dict, model_config_strings = parse_training_results(results_dir)
    models = {parse_model_parameter(x, 'model') for x in model_config_strings}
    best_model, best_model_results, best_model_config_string = get_best_model(models, results_dict)
    # # save the torch model
    # torch.save(best_model, 'trained_model/0.0.2/best_model.pt')
    subimage_handler.model = best_model

    subimage_handler.compute_perceptual_attention('RLS_036_OS_TC', discard_ratio=0.1, notes=figure_notes, normalize_by_subimage=True)
    subimage_handler.compute_perceptual_attention('RLS_036_OS_TC', source_attention=human_attention, discard_ratio=0.1, notes=figure_notes, normalize_by_subimage=True)
    # subimage_handler.compute_perceptual_attention('RLS_036_OS_TC', save_dir=figure_dir, discard_ratio=0.1, notes=figure_notes)
    # subimage_handler.compute_perceptual_attention('9025_OD_2021_widefield_report', source_attention=human_attention, save_dir='figures_example', discard_ratio=0.7,notes=figure_notes)
    # subimage_handler.compute_perceptual_attention('9025_OD_2021_widefield_report', save_dir='figures_example', discard_ratio=0.1,notes=figure_notes)



import torch


def torch_wasserstein_loss(tensor_a,tensor_b):
    #Compute the first Wasserstein distance between two 1D distributions.
    return(torch_cdf_loss(tensor_a,tensor_b,p=1))


def torch_energy_loss(tensor_a,tensor_b):
    # Compute the energy distance between two 1D distributions.
    return((2**0.5)*torch_cdf_loss(tensor_a,tensor_b,p=2))


def torch_cdf_loss(tensor_a,tensor_b,p=1):
    tensor_a = tensor_a / (torch.sum(tensor_a, dim=-1, keepdim=True) + 1e-14)
    tensor_b = tensor_b / (torch.sum(tensor_b, dim=-1, keepdim=True) + 1e-14)
    cdf_tensor_a = torch.cumsum(tensor_a,dim=-1)
    cdf_tensor_b = torch.cumsum(tensor_b,dim=-1)

    # choose different formulas for different norm situations
    if p == 1:
        cdf_distance = torch.sum(torch.abs((cdf_tensor_a-cdf_tensor_b)),dim=-1)
    elif p == 2:
        cdf_distance = torch.sqrt(torch.sum(torch.pow((cdf_tensor_a-cdf_tensor_b),2),dim=-1))
    else:
        cdf_distance = torch.pow(torch.sum(torch.pow(torch.abs(cdf_tensor_a-cdf_tensor_b),p),dim=-1),1/p)

    cdf_loss = cdf_distance.mean()
    return cdf_loss


def torch_validate_distibution(tensor_a,tensor_b):
    if tensor_a.size() != tensor_b.size():
        raise ValueError("Input weight tensors must be of the same size")


def any_image_to_tensor(image, device):
    if type(image) == list or type(image) == tuple:
        image = [[x.to(device) for x in y] for y in image]
    elif type(image) == dict:
        image = {k: [x.to(device) for x in v] for k, v in image.items()}
    else:
        image = image.to(device)
    return image


"""
wget http://www.image-net.org/api/text/imagenet.synset.obtain_synset_list
"""

from pathlib import Path
from nltk.corpus import wordnet

classes_file = 'imagenet.synset.obtain_synset_list'
output_file = 'labels_map_21k.txt'
with open(Path(classes_file)) as f:
    classes = f.read().splitlines()
    classes = [c for c in classes if c != '']
    assert len(classes) in [1000, 21_841]
    classes = [wordnet.synset_from_pos_and_offset('n', int(c[1:])).lemmas()[0].name() for c in classes]
with open(output_file, 'w') as f:
    print('\n'.join(classes), file=f)
    


import cv2
import torch
import numpy as np
import matplotlib.pyplot as plt

from eidl.utils.model_utils import get_trained_model, load_image_preprocess
from eidl.viz.vit_rollout import VITAttentionRollout

# replace the image path to yours
image_path = r'D:\Dropbox\Dropbox\ExpertViT\Datasets\OCTData\oct_v2\reports_cleaned\G_Suspects\RLS_074_OD_TC.jpg'

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

model, image_mean, image_std, image_size, compound_label_encoder = get_trained_model(device, model_param='num-patch-32_image-size-1024-512')

image_normalized, image = load_image_preprocess(image_path, image_size, image_mean, image_std)

# get the prediction
image_tensor = torch.Tensor(image_normalized).unsqueeze(0).to(device)
y_pred, attention_matrix = model(image_tensor, collapse_attention_matrix=False)
predicted_label = np.array([torch.argmax(y_pred).item()])
decoded_label = compound_label_encoder.decode(predicted_label)

print(f'Predicted label: {decoded_label}')

# plot the attention rollout
vit_rollout = VITAttentionRollout(model, device=device, attention_layer_name='attn_drop', head_fusion="mean", discard_ratio=0.5)
rollout = vit_rollout(depth=model.depth, input_tensor=image_tensor)  # rollout on the last layer

rollout_resized = cv2.resize(rollout, dsize=image_size, interpolation=cv2.INTER_LINEAR)
rollout_heatmap = cv2.applyColorMap(cv2.cvtColor((rollout_resized * 255).astype(np.uint8), cv2.COLOR_GRAY2BGR), cv2.COLORMAP_JET)
rollout_heatmap = cv2.cvtColor(rollout_heatmap, cv2.COLOR_BGR2RGB)
alpha = 0.2
output_image = cv2.addWeighted(image.astype(np.uint8), alpha, rollout_heatmap, 1 - alpha, 0)


fig = plt.figure(figsize=(15, 30), constrained_layout=True)
axes = fig.subplots(3, 1)
axes[0].imshow(image.astype(np.uint8))  # plot the original image
axes[0].axis('off')
axes[0].set_title(f'Original image')

axes[1].imshow(rollout_heatmap)  # plot the attention rollout
axes[1].axis('off')
axes[1].set_title(f'Attention rollout')

axes[2].imshow(output_image)  # plot the attention rollout
axes[2].axis('off')
axes[2].set_title(f'Overlayed attention rollout')
plt.show()



"""configs.py - ViT model configurations, based on:
https://github.com/google-research/vision_transformer/blob/master/vit_jax/configs.py
"""

def get_base_config():
    """Base ViT config ViT"""
    return dict(
      dim=768,
      ff_dim=3072,
      num_heads=12,
      num_layers=12,
      attention_dropout_rate=0.0,
      dropout_rate=0.1,
      representation_size=768,
      classifier='token'
    )

def get_b16_config():
    """Returns the ViT-B/16 configuration."""
    config = get_base_config()
    config.update(dict(patches=(16, 16)))
    return config

def get_b32_config():
    """Returns the ViT-B/32 configuration."""
    config = get_b16_config()
    config.update(dict(patches=(32, 32)))
    return config

def get_l16_config():
    """Returns the ViT-L/16 configuration."""
    config = get_base_config()
    config.update(dict(
        patches=(16, 16),
        dim=1024,
        ff_dim=4096,
        num_heads=16,
        num_layers=24,
        attention_dropout_rate=0.0,
        dropout_rate=0.1,
        representation_size=1024
    ))
    return config

def get_l32_config():
    """Returns the ViT-L/32 configuration."""
    config = get_l16_config()
    config.update(dict(patches=(32, 32)))
    return config

def drop_head_variant(config):
    config.update(dict(representation_size=None))
    return config


PRETRAINED_MODELS = {
    'B_16': {
      'config': get_b16_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_16.pth"
    },
    'B_32': {
      'config': get_b32_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_32.pth"
    },
    'L_16': {
      'config': get_l16_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': None
    },
    'L_32': {
      'config': get_l32_config(),
      'num_classes': 21843,
      'image_size': (224, 224),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/L_32.pth"
    },
    'B_16_imagenet1k': {
      'config': drop_head_variant(get_b16_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_16_imagenet1k.pth"
    },
    'B_32_imagenet1k': {
      'config': drop_head_variant(get_b32_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/B_32_imagenet1k.pth"
    },
    'L_16_imagenet1k': {
      'config': drop_head_variant(get_l16_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/L_16_imagenet1k.pth"
    },
    'L_32_imagenet1k': {
      'config': drop_head_variant(get_l32_config()),
      'num_classes': 1000,
      'image_size': (384, 384),
      'url': "https://github.com/lukemelas/PyTorch-Pretrained-ViT/releases/download/0.0.2/L_32_imagenet1k.pth"
    },
}



import copy

from graphviz import Digraph
import torch
from torch.autograd import Variable, Function

def iter_graph(root, callback):
    queue = [root]
    seen = set()
    while queue:
        fn = queue.pop()
        if fn in seen:
            continue
        seen.add(fn)
        for next_fn, _ in fn.next_functions:
            if next_fn is not None:
                queue.append(next_fn)
        callback(fn)

def is_bad_grad(grad_output):
    if grad_output is None:
        return False
    return grad_output.isnan().any() or (grad_output.abs() >= 1e6).any()

def register_hooks(var, name):
    fn_dict = {}
    def hook_cb(fn):
        def register_grad(grad_input, grad_output):
            fn_dict[fn] = copy.deepcopy(grad_input) # HERE
        fn.register_hook(register_grad)
    iter_graph(var.grad_fn, hook_cb)
    def make_dot():
        node_attr = dict(style='filled',
                        shape='box',
                        align='left',
                        fontsize='12',
                        ranksep='0.1',
                        height='0.2')
        dot = Digraph(name, node_attr=node_attr, graph_attr=dict(size="12,12"))

        def size_to_str(size):
            return '('+(', ').join(map(str, size))+')'

        def build_graph(fn):
            if hasattr(fn, 'variable'):  # if GradAccumulator
                u = fn.variable
                node_name = 'Variable\n ' + size_to_str(u.size())
                dot.node(str(id(u)), node_name, fillcolor='lightblue')
            else:
                assert fn in fn_dict, fn
                fillcolor = 'white'
                if any(is_bad_grad(gi) for gi in fn_dict[fn]):
                    fillcolor = 'red'
                dot.node(str(id(fn)), str(type(fn).__name__), fillcolor=fillcolor)
            for next_fn, _ in fn.next_functions:
                if next_fn is not None:
                    next_id = id(getattr(next_fn, 'variable', next_fn))
                    dot.edge(str(next_id), str(id(fn)))
        iter_graph(var.grad_fn, build_graph)

        return dot

    return make_dot


import torch
from torch import nn
from torch import Tensor
from PIL import Image
from torchvision.transforms import Compose, Resize, ToTensor
from einops import rearrange, reduce, repeat
from einops.layers.torch import Rearrange, Reduce
from torchsummary import summary
import torch.nn.functional as F


class PatchEmbedding(nn.Module):
    def __init__(self, img_size: int, in_channels: int = 3, patch_size: int = 16, embedding_size: int = 768,
                 mode="conv"):
        """
        patch embedding mode can be either conv or linear

        this implementation is inpired by
        https://github.com/lucidrains/vit-pytorch/blob/main/vit_pytorch/vit.py
        https://towardsdatascience.com/implementing-visualttransformer-in-pytorch-184f9f16f632
        :param in_channels:
        :param patch_size:
        :param embedding_size:
        :param mode:
        """
        self.patch_size = patch_size
        super().__init__()
        if mode == 'linear':
            self.patch_embedding = nn.Sequential(
                # break-down the image in s1 x s2 patches and flat them
                Rearrange('b c (h s1) (w s2) -> b (h w) (s1 s2 c)', s1=patch_size, s2=patch_size),
                nn.Linear(patch_size * patch_size * in_channels, embedding_size)
            )
        elif mode == 'conv':
            self.patch_embedding = nn.Sequential(
                # using a conv layer instead of a linear one -> performance gains
                nn.Conv2d(in_channels, embedding_size, kernel_size=patch_size, stride=patch_size),
                Rearrange('b e (h) (w) -> b (h w) e'),
            )
        else:
            raise Exception("Unsupported patch embedding mode, please use conv or linear.")
        self.cls_token = nn.Parameter(torch.randn(1, 1, embedding_size))
        self.positions = nn.Parameter(torch.randn((img_size // patch_size) ** 2 + 1, embedding_size))

    def forward(self, x: Tensor) -> Tensor:
        b, _, _, _ = x.shape

        x = self.patch_embedding(x)
        cls_tokens = repeat(self.cls_token, '() n e -> b n e', b=b)

        x = torch.cat([cls_tokens, x], dim=1)
        x += self.positions

        return x


class MultiHeadAttention(nn.Module):
    def __init__(self, emb_size: int = 512, num_heads: int = 8, dropout: float = 0):
        super().__init__()
        self.emb_size = emb_size
        self.num_heads = num_heads
        # self.keys = nn.Linear(emb_size, emb_size)
        # self.queries = nn.Linear(emb_size, emb_size)
        # self.values = nn.Linear(emb_size, emb_size)
        self.qkv = nn.Linear(emb_size, emb_size * 3)

        self.att_drop = nn.Dropout(dropout)
        self.projection = nn.Linear(emb_size, emb_size)

    def forward(self, x: Tensor, mask: Tensor = None) -> Tensor:
        # split keys, queries and values in num_heads
        # queries = rearrange(self.queries(x), "b n (h d) -> b h n d", h=self.num_heads)
        # keys = rearrange(self.keys(x), "b n (h d) -> b h n d", h=self.num_heads)
        # values = rearrange(self.values(x), "b n (h d) -> b h n d", h=self.num_heads)
        qkv = rearrange(self.qkv(x), "b n (h d qkv) -> (qkv) b h n d", h=self.num_heads, qkv=3)
        queries, keys, values = qkv[0], qkv[1], qkv[2]

        # sum up over the last axis
        energy = torch.einsum('bhqd, bhkd -> bhqk', queries, keys)  # batch, num_heads, query_len, key_len
        if mask is not None:
            fill_value = torch.finfo(torch.float32).min
            energy.mask_fill(~mask, fill_value)

        scaling = self.emb_size ** (1 / 2)
        att = F.softmax(energy, dim=-1) / scaling
        att = self.att_drop(att)
        # sum up over the third axis
        out = torch.einsum('bhal, bhlv -> bhav ', att, values)
        out = rearrange(out, "b h n d -> b n (h d)")
        out = self.projection(out)
        return out


class ResidualAdd(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        res = x
        x = self.fn(x, **kwargs)
        x += res
        return x


class FeedForwardBlock(nn.Sequential):
    def __init__(self, emb_size: int, expansion: int = 4, drop_p: float = 0.):
        super().__init__(
            nn.Linear(emb_size, expansion * emb_size),
            nn.GELU(),  # the Gaussian Error Linear Units function
            nn.Dropout(drop_p),
            nn.Linear(expansion * emb_size, emb_size),
        )


class TransformerEncoderBlock(nn.Sequential):
    def __init__(self,
                 emb_size: int = 768,
                 drop_p: float = 0.,
                 forward_expansion: int = 4,
                 forward_drop_p: float = 0.,
                 **kwargs):
        super().__init__(
            ResidualAdd(nn.Sequential(
                nn.LayerNorm(emb_size),
                MultiHeadAttention(emb_size, **kwargs),
                nn.Dropout(drop_p)
            )),
            ResidualAdd(nn.Sequential(
                nn.LayerNorm(emb_size),
                FeedForwardBlock(
                    emb_size, expansion=forward_expansion, drop_p=forward_drop_p),
                nn.Dropout(drop_p)
            )
            ))


class TransformerEncoder(nn.Sequential):
    def __init__(self, depth: int = 12, **kwargs):
        super().__init__(*[TransformerEncoderBlock(**kwargs) for _ in range(depth)])


class ClassificationHead(nn.Sequential):
    def __init__(self, n_classes, emb_size: int = 768):
        super().__init__(
            Reduce('b n e -> b e', reduction='mean'),
            nn.LayerNorm(emb_size),
            nn.Linear(emb_size, n_classes))


class ViT(nn.Sequential):
    def __init__(self,
                 n_classes: int,
                 img_size: int,
                 in_channels: int = 3,
                 patch_size: int = 16,
                 emb_size: int = 768,
                 depth: int = 12,
                 **kwargs):
        super().__init__()
        self.patch_embedding = PatchEmbedding(img_size, in_channels, patch_size, emb_size)
        self.transformer_encoder = TransformerEncoder(depth=depth, emb_size=emb_size, **kwargs)
        self.classification_head = ClassificationHead(n_classes, emb_size)

    def forward(self, x):
        x = self.patch_embedding(x)
        x = self.transformer_encoder(x)
        return F.softmax(self.classification_head(x))



import os.path

import numpy as np
import timm
import torch
from matplotlib import pyplot as plt

from eidl.datasets.ODIRDatasets import get_ODIRDataset
from eidl.utils.training_utils import train

# from Models.ViT import ViT
# from Models.ViT_luci import ViT_luci
# from pytorch_pretrained_vit import ViT

# Change the following to the file path on your system #########
data_root = 'D:/Dropbox/Dropbox/ExpertViT/Datasets/ODIR'
data_export_root = 'C:/Data/ODIR'
results_dir = 'archived/results_odir'

n_jobs = 16  # n jobs for loading data from hard drive

# generic training parameters ##################################
epochs = 500
random_seed = 42
batch_size = 8
folds = 3

# grid search hyper-parameters ##################################
################################################################
depths = 1, 3
# depths = 3,

################################################################
# alphas = 0.0, 1e-2, 0.1, 0.25, 0.5, 0.75, 1.0
alphas = 1e-2, 0.0
# alphas = .0,

################################################################
# lrs = 1e-2, 1e-3, 1e-4
lrs = 1e-3, 1e-4
non_pretrained_lr_scaling = 1e-2

################################################################
aoi_loss_distance_types = 'Wasserstein', 'cross-entropy'

################################################################
# model_names = 'base', 'vit_small_patch32_224_in21k', 'vit_small_patch16_224_in21k', 'vit_large_patch16_224_in21k'
# model_names = 'base', 'vit_small_patch32_224_in21k'
# model_names = 'vit_base_patch32_384',
model_names = 'vit_large_patch32_384',
pretrained = True
# model_names = 'base',

train_size = 0.8

model_name = 'ViTNotPretrained'
save_dir = f'results-08_15_2023_15_43_28/SavedModels/{model_names[0]}-Pretrained_{pretrained}/'

if __name__ == '__main__':
    save_dir = f'{results_dir}/{model_name}/'

    # normalize_mean_resnet = [0.485, 0.456, 0.406]
    # normalize_std_resnet = [0.229, 0.224, 0.225]
    # normalize_mean_ViTPretrained = [0.5, 0.5, 0.5]
    # normalize_std_ViTPretrained = [0.5, 0.5, 0.5]

    torch.manual_seed(random_seed)
    np.random.seed(random_seed)

    if not os.path.exists(save_dir): os.mkdir(save_dir)

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    # model = ViT('B_16_imagenet1k', pretrained=True, num_classes=num_classes, image_size=512).to(device)  # pretrained ViT

    # config = resolve_data_config({}, model=model)
    # transform = create_transform(**config)

    train_data_loader, val_data_loader, train_size, val_size, input_shape, num_classes = get_ODIRDataset(data_root, train_size, batch_size, data_export_root=data_export_root)
    model = timm.create_model(model_names[0], pretrained=pretrained, num_classes=8, img_size=input_shape[-1]).to(device)  # weights from 'https://storage.googleapis.com/vit_models/augreg/L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1.npz', official Google JAX implementation
    # model = ViT_LSTM(image_size=input_shape[1:], num_patches=16, num_classes=8, embed_dim=512, dim_head=512, depth=6, heads=8, mlp_dim=2048, weak_interaction=False).to(device)  # weights from 'https://storage.googleapis.com/vit_models/augreg/L_16-i21k-300ep-lr_0.001-aug_medium1-wd_0.1-do_0.1-sd_0.1.npz', official Google JAX implementation

    # for name, p in model.named_parameters():
    #     # print(f'Layer {name}, grad: {p.requires_grad}')
    #     p.requires_grad = False
    # for p in model.fc_norm.parameters():
    #     p.requires_grad = True
    # for p in model.head.parameters():
    #     p.requires_grad = True

    # print("Model Summary: ")
    # summary(model, input_size=input_shape)

    # optimizer = torch.optim.SGD(model.parameters(), momentum=0.9, nesterov=True, lr=1e-4, weight_decay=0.)  # parameter used in https://github.com/Zoe0123/Vision-Transformer-for-Chest-X-Ray-Classification/blob/main/vit.ipynb
    optimizer = torch.optim.Adam(model.parameters(), lr=lrs[0])
    # criterion = nn.CrossEntropyLoss(label_smoothing=0.1)
    # criterion = LabelSmoothingCrossEntropy(smoothing=0.1).cuda()  # from here: https://timm.fast.ai/
    # optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    # optimizer = torch.optim.SGD(model.parameters(), momentum=0.9, nesterov=True, lr=0.01, weight_decay=0.)  # parameter used in https://github.com/Zoe0123/Vision-Transformer-for-Chest-X-Ray-Classification/blob/main/vit.ipynb
    # optimizer = torch.optim.SGD(model.parameters(), lr=lr, momentum=0.9, weight_decay=1e-4)
    # scheduler = lr_scheduler.StepLR(optimizer, step_size=7, gamma=0.1)

    training_history = train(model, optimizer, train_data_loader, val_data_loader, epochs, model_name, save_dir, n_classes=8)

    plt.plot(training_history['loss_train'])
    plt.plot(training_history['loss_val'])
    plt.show()


import torchvision
from torch import nn
import torch.nn.functional as F


class ResNet(nn.Sequential):
    def __init__(self,
                 n_classes: int,
                 **kwargs):
        super().__init__()
        self.resnet = torchvision.models.resnet34(weights=torchvision.models.ResNet34_Weights)
        self.resnet.fc = nn.Linear(self.resnet.fc.in_features, n_classes)

    def forward(self, x):
        x = self.resnet(x)
        return F.softmax(x, dim=1)



"""
Adapted from https://github.com/lukemelas/simple-bert
"""
 
import numpy as np
from torch import nn
from torch import Tensor 
from torch.nn import functional as F


def split_last(x, shape):
    "split the last dimension to given shape"
    shape = list(shape)
    assert shape.count(-1) <= 1
    if -1 in shape:
        shape[shape.index(-1)] = int(x.size(-1) / -np.prod(shape))
    return x.view(*x.size()[:-1], *shape)


def merge_last(x, n_dims):
    "merge the last n_dims to a dimension"
    s = x.size()
    assert n_dims > 1 and n_dims < len(s)
    return x.view(*s[:-n_dims], -1)


class MultiHeadedSelfAttention(nn.Module):
    """Multi-Headed Dot Product Attention"""
    def __init__(self, dim, num_heads, dropout):
        super().__init__()
        self.proj_q = nn.Linear(dim, dim)
        self.proj_k = nn.Linear(dim, dim)
        self.proj_v = nn.Linear(dim, dim)
        self.drop = nn.Dropout(dropout)
        self.n_heads = num_heads
        self.scores = None # for visualization

    def forward(self, x, mask):
        """
        x, q(query), k(key), v(value) : (B(batch_size), S(seq_len), D(dim))
        mask : (B(batch_size) x S(seq_len))
        * split D(dim) into (H(n_heads), W(width of head)) ; D = H * W
        """
        # (B, S, D) -proj-> (B, S, D) -split-> (B, S, H, W) -trans-> (B, H, S, W)
        q, k, v = self.proj_q(x), self.proj_k(x), self.proj_v(x)
        q, k, v = (split_last(x, (self.n_heads, -1)).transpose(1, 2) for x in [q, k, v])
        # (B, H, S, W) @ (B, H, W, S) -> (B, H, S, S) -softmax-> (B, H, S, S)
        scores = q @ k.transpose(-2, -1) / np.sqrt(k.size(-1))
        if mask is not None:
            mask = mask[:, None, None, :].float()
            scores -= 10000.0 * (1.0 - mask)
        scores = self.drop(F.softmax(scores, dim=-1))
        # (B, H, S, S) @ (B, H, S, W) -> (B, H, S, W) -trans-> (B, S, H, W)
        h = (scores @ v).transpose(1, 2).contiguous()
        # -merge-> (B, S, D)
        h = merge_last(h, 2)
        self.scores = scores
        return h


class PositionWiseFeedForward(nn.Module):
    """FeedForward Neural Networks for each position"""
    def __init__(self, dim, ff_dim):
        super().__init__()
        self.fc1 = nn.Linear(dim, ff_dim)
        self.fc2 = nn.Linear(ff_dim, dim)

    def forward(self, x):
        # (B, S, D) -> (B, S, D_ff) -> (B, S, D)
        return self.fc2(F.gelu(self.fc1(x)))


class Block(nn.Module):
    """Transformer Block"""
    def __init__(self, dim, num_heads, ff_dim, dropout):
        super().__init__()
        self.attn = MultiHeadedSelfAttention(dim, num_heads, dropout)
        self.proj = nn.Linear(dim, dim)
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        self.pwff = PositionWiseFeedForward(dim, ff_dim)
        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.drop = nn.Dropout(dropout)

    def forward(self, x, mask):
        h = self.drop(self.proj(self.attn(self.norm1(x), mask)))
        x = x + h
        h = self.drop(self.pwff(self.norm2(x)))
        x = x + h
        return x


class Transformer(nn.Module):
    """Transformer with Self-Attentive Blocks"""
    def __init__(self, num_layers, dim, num_heads, ff_dim, dropout):
        super().__init__()
        self.blocks = nn.ModuleList([
            Block(dim, num_heads, ff_dim, dropout) for _ in range(num_layers)])

    def forward(self, x, mask=None):
        for block in self.blocks:
            x = block(x, mask)
        return x



import torch

import numpy as np

from eidl.Models.ExpertAttentionViT import Attention as ExpertAttentionViTAttention

from timm.models.vision_transformer import VisionTransformer

from eidl.utils.torch_utils import any_image_to_tensor


def rollout(depth, grid_size, attentions, discard_ratio, head_fusion, normalize=True, return_raw_attention=False, *args, **kwargs):
    """

    Parameters
    ----------
    depth
    grid_size
    attentions
    discard_ratio
    head_fusion
    return_raw_attention: if false, the classification token's attention will be normalized and returned as a mask
                          if true, the raw attention will be returned
    normalize: if true, the attention will be normalized by dividing by the maximum attention
    args
    kwargs

    Returns
    -------

    """
    result = torch.eye(attentions[0].size(-1))
    with torch.no_grad():
        for i, attention in enumerate(attentions):
            if head_fusion == "mean":
                attention_heads_fused = attention.mean(axis=1)
            elif head_fusion == "max":
                attention_heads_fused = attention.max(axis=1)[0]
            elif head_fusion == "min":
                attention_heads_fused = attention.min(axis=1)[0]
            else:
                raise "Attention head fusion type Not supported"

            # Drop the lowest attentions, but
            # don't drop the class token
            flat = attention_heads_fused.view(attention_heads_fused.size(0), -1)
            _, indices = flat.topk(int(flat.size(-1) * discard_ratio), -1, False)
            indices = indices[indices != 0]
            flat[0, indices] = 0

            I = torch.eye(attention_heads_fused.size(-2), attention_heads_fused.size(-1))  # TODO check why this is not square for ExpertAttentionModel
            a = (attention_heads_fused + 1.0 * I) / 2
            a = a / a.sum(dim=-1)

            result = torch.matmul(a, result)
            if i == depth:
                break

    if not return_raw_attention:
        result = result[0, 0, 1:]  # Look at the total attention between the class token, # and the image patches
    else:
        result = result[0]
    if normalize:
        result = result / torch.max(result)
    return result.numpy()

class VITAttentionRollout:
    def __init__(self, model, device, attention_layer_name='attn_drop', head_fusion="mean",
                 discard_ratio=0.9, *args, **kwargs):
        """


        Parameters
        ----------
        model
        device
        attention_layer_name
        head_fusion
        discard_ratio: discard the bottom x% of the attention
        args
        kwargs
        """
        self.model = model
        self.device = device
        self.head_fusion = head_fusion
        self.discard_ratio = discard_ratio

        self.attention_layer_count = 0
        for name, module in self.model.named_modules():
            if attention_layer_name in name or isinstance(module, ExpertAttentionViTAttention):
                module.register_forward_hook(self.get_attention)
                self.attention_layer_count += 1
        if self.attention_layer_count == 0:
            raise ValueError("No attention layer in the given model")
        if self.attention_layer_count != self.model.depth:
            raise ValueError(f"Model depth ({self.model.depth}) does not match attention layer count {self.attention_layer_count}")
        self.attentions = []

    def get_attention(self, module, input, output):
        if isinstance(module, ExpertAttentionViTAttention):
            # attention_output = rearrange(output[0], 'b n (h d) -> b h n d', h=1)  # TODO remove the hardcoding num heads
            attention_output = output[1]
        else:
            attention_output = output
        self.attentions.append(attention_output.cpu())

    def __call__(self, depth, in_data, *args, **kwargs):
        if depth > self.attention_layer_count:
            raise ValueError(f"Given depth ({depth}) is greater than the number of attenion layers in the model ({self.attention_layer_count})")
        self.attentions = []

        output = self.model(in_data, *args, **kwargs)
        # if isinstance(self.model, VisionTransformer):
        #     output = self.model(in_data)
        # elif fix_sequence is not None:
        #
        # else:
        #     output = self.model(in_data)
        # with torch.no_grad():
        #     if isinstance(self.model, ViT_LSTM):
        #         x = self.model.to_patch_embedding(input_tensor.to(self.device))
        #         b, n, _ = x.shape
        #
        #         cls_tokens = repeat(self.model.cls_token, '1 1 d -> b 1 d', b=b)
        #         x = torch.cat((cls_tokens, x), dim=1)
        #         x += self.model.pos_embedding[:, :(n + 1)]
        #         x = self.model.dropout(x)
        #
        #         # rollout the call: x, att_matrix = self.transformer(x)
        #         for i, (attn, ff) in enumerate(self.model.transformer.layers):
        #             out, attention = attn(x)
        #             x = out + x
        #             x = ff(x) + x
        #             if i + 1 == layer:
        #                 break
        #         # TODO
        #     elif isinstance(self.model, ExpertTimmVisionTransformer):
        #         x = best_model.vision_transformer.patch_embed(tensor.to(device))
        #         x = best_model.vision_transformer._pos_embed(x)
        #         x = best_model.vision_transformer.norm_pre(x)
        #
        #         for i in range(layer):  # iterate to before the designated layer
        #             x = best_model.vision_transformer.blocks[i](x)
        #
        #         this_attention = best_model.vision_transformer.blocks[
        #             layer].attn  # TODO rewrite to directly retrieve attention activation from the target block, using the ExtensionModel's redefined block
        #         B, N, C = x.shape
        #         qkv = this_attention.attention.qkv(x).reshape(B, N, 3, this_attention.attention.num_heads,
        #                                                       C // this_attention.attention.num_heads).permute(
        #             2, 0, 3, 1, 4)
        #         q, k, v = qkv.unbind(0)
        #
        #         this_q = q[0, head, 1:, channel]
        #         this_k = k[0, head, 1:, channel]
        #         query = torch.sigmoid(this_q).reshape(grid_size).cpu().detach().numpy()
        #         key = torch.sigmoid(this_k).reshape(grid_size).cpu().detach().numpy()
        #
        #         activation = q[:, head, :, channel] * k[:, head, :, channel].T
        #         class_activation = activation[0, 1:].reshape(grid_size).cpu().detach().numpy()
        #     elif isinstance(self.model, VisionTransformer):
        #         output = self.model(input_tensor)

        return rollout(depth, self.model.get_grid_size(), self.attentions, self.discard_ratio, self.head_fusion, *args, **kwargs)


"""
This file implements ViT using the cropped images


"""

import itertools
import os
import pickle
from datetime import datetime

import numpy as np
import torch
from torch import optim, nn
from torch.utils.data import DataLoader

from eidl.datasets.OCTDataset import get_oct_test_train_val_folds
from eidl.utils.iter_utils import collate_fn
from eidl.utils.model_utils import get_model
from eidl.utils.training_utils import train_oct_model, get_class_weight

# User parameters ##################################################################################

# Change the following to the file path on your system #########
# data_root = 'D:/Dropbox/Dropbox/ExpertViT/Datasets/OCTData/oct_v2'
data_root = r'C:\Dropbox\ExpertViT\Datasets\OCTData\oct_v2'
# data_root = r'C:\Users\apoca_vpmhq3c\Dropbox\ExpertViT\Datasets\OCTData\oct_v2'

cropped_image_data_path = r'C:\Dropbox\ExpertViT\Datasets\OCTData\oct_v2\oct_reports_info.p'
# results_dir = 'results'
# use_saved_folds = 'results-01_07_2024_10_53_56'

results_dir = '../temp/results'
use_saved_folds = '../temp/results-01_07_2024_15_13_23'

n_jobs = 20  # n jobs for loading data from hard drive and z-norming the subimages

# generic training parameters ##################################
epochs = 15
random_seed = 42
batch_size = 2
folds = 3

# grid search hyper-parameters ##################################
################################################################
# depths = 1, 3
depths = 1,

################################################################
# alphas = 0.0, 1e-2, 0.1, 0.25, 0.5, 0.75, 1.0
# alphas = 1e-2, 0.0
alphas = 1e-2,
# alphas = .0,

################################################################
# lrs = 1e-2, 1e-3, 1e-4
# lrs = 1e-4, 1e-5
# lrs = 1e-4,
lrs = 1e-3,

non_pretrained_lr_scaling = 1e-2

################################################################
# aoi_loss_distance_types = 'Wasserstein', 'cross-entropy'
aoi_loss_distance_types = 'cross-entropy',

################################################################
# model_names = 'base', 'vit_small_patch32_224_in21k', 'vit_small_patch16_224_in21k', 'vit_large_patch16_224_in21k'
# model_names = 'base', 'vit_small_patch32_224_in21k'
model_names = 'vit_small_patch32_224_in21k_subimage',
# model_names = 'base_subimage',
# model_names = 'inception_v4_subimage'

grid_search_params = {
    'vit_small_patch32_224_in21k_subimage': {
        'alphas': 1e-2,
        'lrs': 1e-3,
        'aoi_loss_distance_types': 'cross-entropy',
        'optimizer': optim.SGD,
    },

    'base_subimage': {
        'alphas': 1e-2,
        'lrs': 1e-4,
        'aoi_loss_distance_types': 'cross-entropy',
        'depths': 1,
        'optimizer': optim.Adam,},

    'inception_v4_subimage': {
        'lrs': 1e-3,
        'optimizer': optim.SGD,
    }
}



################################################################
image_size = 1024, 512
patch_size = 32, 32
gaussian_smear_sigma = 0.5

# end of user parameters #############################################################################
if __name__ == '__main__':

    torch.manual_seed(random_seed)
    np.random.seed(random_seed)

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    if use_saved_folds:
        print(f"Using saved folds from {use_saved_folds}")
        folds = pickle.load(open(os.path.join(use_saved_folds, 'folds.p'), 'rb'))
        test_dataset = pickle.load(open(os.path.join(use_saved_folds, 'test_dataset.p'), 'rb'))
        image_stats = pickle.load(open(os.path.join(use_saved_folds, 'image_stats.p'), 'rb'))
        test_dataset.compound_label_encoder = pickle.load(open(os.path.join(use_saved_folds, 'compound_label_encoder.p'), 'rb'))
        results_dir = use_saved_folds
    else:
        print("Creating data set")
        folds, test_dataset, image_stats = get_oct_test_train_val_folds(data_root, image_size=image_size, n_folds=folds, n_jobs=n_jobs,
                                                                                    cropped_image_data_path=cropped_image_data_path,
                                                                                    patch_size=patch_size, gaussian_smear_sigma=gaussian_smear_sigma)
        now = datetime.now()
        dt_string = now.strftime("%m_%d_%Y_%H_%M_%S")
        results_dir = f"{results_dir}-{dt_string}"
        if not os.path.isdir(results_dir):
            os.mkdir(results_dir)
            print(f"Results will be save to {results_dir}")
        else:
            print(f"Results exist in {results_dir}, overwritting the results")
        pickle.dump(folds, open(os.path.join(results_dir, 'folds.p'), 'wb'))
        pickle.dump(test_dataset, open(os.path.join(results_dir, 'test_dataset.p'), 'wb'))
        pickle.dump(image_stats, open(os.path.join(results_dir, 'image_stats.p'), 'wb'))
        pickle.dump(test_dataset.compound_label_encoder, open(os.path.join(results_dir, 'compound_label_encoder.p'), 'wb'))

    train_trial_dataset, valid_dataset, train_unique_img_dataset = folds[0]  # TODO using only one fold for now
    # test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)  # TODO use the test loader in the future

    # process the grid search parameters
    parameters = set()
    # for model, grid_search_params


    for depth, alpha, model_name, lr, aoi_loss_dist in itertools.product(depths, alphas, model_names, lrs, aoi_loss_distance_types):
        # if model_name == 'pretrained':
        #     this_lr = lr * non_pretrained_lr_scaling
        #     this_depth = None  # depth does not affect the pretrained model
        # else:
        this_lr = lr
        this_depth = depth
        if 'inception' in model_name:  # inception net doesn't have depth and alpha
            this_params = (model_name, None, None, None, this_lr)
        else:
            this_params = (model_name, this_depth, alpha, aoi_loss_dist, this_lr)
        parameters.add(this_params)

    for i, parameter in enumerate(parameters):  # iterate over the grid search parameters
        model_name, depth, alpha, aoi_loss_dist, lr = parameter
        model, grid_size = get_model(model_name, image_size=image_stats['subimage_sizes'], depth=depth, device=device, patch_size=patch_size)
        model_config_string = f'model-{model_name}_alpha-{alpha}_dist-{aoi_loss_dist}_depth-{model.depth}_lr-{lr}'
        print(f"Grid search [{i}] of {len(parameters)}: {model_config_string}")

        if 'inception' in model_name:
            train_dataset = train_unique_img_dataset
        else:
            train_dataset = train_trial_dataset
        train_dataset.create_aoi(grid_size=grid_size, use_subimages=True)
        valid_dataset.create_aoi(grid_size=grid_size, use_subimages=True)

        class_weights = get_class_weight(train_dataset.labels_encoded, 2).to(device)

        # optimizer = optim.Adam(model.parameters(), lr=lr)
        optimizer = optim.SGD(model.parameters(), lr=lr)
        criterion = nn.CrossEntropyLoss()

        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
        valid_loader = DataLoader(valid_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

        train_loss_list, train_acc_list, valid_loss_list, valid_acc_list = train_oct_model(
            model, model_config_string, train_loader, valid_loader, results_dir=results_dir, optimizer=optimizer, num_epochs=epochs,
            alpha=alpha, dist=aoi_loss_dist, l2_weight=None, class_weights=class_weights)

    # viz_oct_results(results_dir, test_image_path, test_image_main, batch_size, image_size, n_jobs=n_jobs)


import itertools
import os
import pickle
from datetime import datetime

import numpy as np
import torch
from torch import optim, nn
from torch.utils.data import DataLoader

from eidl.datasets.OCTDataset import get_oct_test_train_val_folds
from eidl.utils.iter_utils import collate_fn
from eidl.utils.model_utils import get_model
from eidl.utils.training_utils import train_oct_model

# User parameters ##################################################################################

# Change the following to the file path on your system #########
data_root = 'D:/Dropbox/Dropbox/ExpertViT/Datasets/OCTData/oct_v2'
results_dir = 'results'

n_jobs = 16  # n jobs for loading data from hard drive

# generic training parameters ##################################
epochs = 100
random_seed = 42
batch_size = 8
folds = 3

# grid search hyper-parameters ##################################
################################################################
# depths = 1, 3
depths = 1,

################################################################
# alphas = 0.0, 1e-2, 0.1, 0.25, 0.5, 0.75, 1.0
# alphas = 1e-2, 0.0
alphas = 1e-2,
# alphas = .0,

################################################################
# lrs = 1e-2, 1e-3, 1e-4
# lrs = 1e-4, 1e-5
# lrs = 1e-4,
lrs = 1e-4,

non_pretrained_lr_scaling = 1e-2

################################################################
# aoi_loss_distance_types = 'Wasserstein', 'cross-entropy'
aoi_loss_distance_types = 'cross-entropy',

################################################################
# model_names = 'base', 'vit_small_patch32_224_in21k', 'vit_small_patch16_224_in21k', 'vit_large_patch16_224_in21k'
# model_names = 'base', 'vit_small_patch32_224_in21k'
# model_names = 'vit_small_patch32_224_in21k',
model_names = 'base',

################################################################
image_size = 1024, 512

# end of user parameters #############################################################################
if __name__ == '__main__':

    torch.manual_seed(random_seed)
    np.random.seed(random_seed)

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    now = datetime.now()
    dt_string = now.strftime("%m_%d_%Y_%H_%M_%S")
    results_dir = f"{results_dir}-{dt_string}"
    if not os.path.isdir(results_dir):
        os.mkdir(results_dir)
        print(f"Results will be save to {results_dir}")
    else:
        print(f"Results exist in {results_dir}, overwritting the results")

    # TODO train and test should use the same std and mean to normalize, moreover train and test should splitted during run time
    print("Creating data set")
    folds, test_dataset, image_means, image_stds = get_oct_test_train_val_folds(data_root, image_size=image_size, n_folds=folds, n_jobs=n_jobs)
    pickle.dump((image_means, image_stds), open(os.path.join(results_dir, 'image_means_stds.p'), 'wb'))
    pickle.dump(test_dataset.compound_label_encoder, open(os.path.join(results_dir, 'compound_label_encoder.p'), 'wb'))
    # test_dataset = get_oct_dataset(test_image_path, test_image_main, image_size=image_size, n_jobs=n_jobs)

    train_dataset, valid_dataset = folds[0]  # TODO using only one fold for now

    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    valid_loader = DataLoader(valid_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)
    # test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

    # save the data loader # TODO use test and save folds in the future
    pickle.dump(folds, open(os.path.join(results_dir, 'folds.p'), 'wb'))
    pickle.dump(test_dataset, open(os.path.join(results_dir, 'test_dataset.p'), 'wb'))

    parameters = set()
    for depth, alpha, model_name, lr, aoi_loss_dist in itertools.product(depths, alphas, model_names, lrs, aoi_loss_distance_types):
        if model_name != 'base':
            this_lr = lr * non_pretrained_lr_scaling
            this_depth = None  # depth does not affect the pretrained model
        else:
            this_lr = lr
            this_depth = depth
        parameters.add((this_depth, alpha, model_name, this_lr, aoi_loss_dist))

    for i, parameter in enumerate(parameters):  # iterate over the grid search parameters
        depth, alpha, model_name, lr, aoi_loss_dist = parameter
        model, grid_size = get_model(model_name, image_size=image_size, depth=depth, device=device)
        model_config_string = f'model-{model_name}_alpha-{alpha}_dist-{aoi_loss_dist}_depth-{model.depth}_lr-{lr}'
        print(f"Grid search [{i}] of {len(parameters)}: {model_config_string}")

        train_dataset.create_aoi(grid_size)
        valid_dataset.create_aoi(grid_size)

        optimizer = optim.Adam(model.parameters(), lr=lr)

        criterion = nn.CrossEntropyLoss()
        train_loss_list, train_acc_list, valid_loss_list, valid_acc_list = train_oct_model(
            model, model_config_string, train_loader, valid_loader, results_dir=results_dir, optimizer=optimizer, num_epochs=epochs,
            alpha=alpha, dist=aoi_loss_dist, l2_weight=None)

    # viz_oct_results(results_dir, test_image_path, test_image_main, batch_size, image_size, n_jobs=n_jobs)


import sys
import os
import pandas as pd


def load_oct_data(data_dir, expert_data_root='ExpertData', experiment_image_root='ExpIms'):
    expert_data_root = os.path.join(data_dir, expert_data_root)
    experiment_image_root = os.path.join(data_dir, experiment_image_root)

    for experiment_days in os.listdir(expert_data_root):  # iterate over the days
        experiment_days_dir = os.path.join(expert_data_root, experiment_days)

        for experiment in os.listdir(experiment_days_dir):
            experiment_dir = os.path.join(experiment_days_dir, experiment)
            for expert in os.listdir(experiment_dir):
                expert_dir = os.path.join(experiment_dir, expert)
                for image_name in os.listdir(expert_dir):
                    image_dir = os.path.join(expert_dir, image_name)
                    image_path = os.path.join(experiment_image_root, image_name)
                    for fixation in os.listdir(image_dir):
                        fixation_dir = os.path.join(image_dir, fixation)
                        fixation_path = os.path.join(image_path, fixation)
                        df = pd.read_csv(fixation_dir)
                        sequences = np.array(df[['norm_pos_x', 'norm_pos_y']])
                        sequences[:, 1] = 1 - sequences[:, 1]
                        yield image_path, sequences



if __name__ == '__main__':
    data_root = 'D:/Dropbox/ExpertViT/Datasets/OCTData'




import os.path

import cv2
import numpy as np
from matplotlib import pyplot as plt

from eidl.utils.iter_utils import collate_fn, collate_subimages
from eidl.utils.image_utils import preprocess_subimages, z_norm_subimages, process_aoi, patchify
from eidl.utils.torch_utils import any_image_to_tensor
from eidl.viz.vit_rollout import VITAttentionRollout
from eidl.viz.viz_oct_results import register_cmap_with_alpha
from eidl.viz.viz_utils import plot_subimage_rolls, plot_image_attention


class SubimageHandler:

    def __init__(self):
        self.subimage_mean = None
        self.subimage_std = None
        self.image_data_dict = None
        self.model = None
        self.attention_cache = {}  # holds the attention for each image, so that it does not have to be recomputed


    def load_image_data(self, image_data_dict, *args, **kwargs):
        """

        Parameters
        ----------
        image_data_dict: dict
            image_name: str: image names are the keys of the dict
                'image': np.array: the original image
                'sub_images': dict
                    'En-face_52.0micrometer_Slab_(Retina_View)': str
                        'sub_image': np.array
                        'position': list of four two-int tuples
                    'Circumpapillary_RNFL':                             same as above
                    'RNFL_Thickness_(Retina_View)':                     same as above
                    'GCL_Thickness_(Retina_View)':                      same as above
                    'RNFL_Probability_and_VF_Test_points(Field_View)':  same as above
                    'GCL+_Probability_and_VF_Test_points':              same as above
                'label': str: 'G', 'S', 'G_Suspects', 'S_Suspects'

        Returns
        -------
        dict:
            image_name: str: image names are the keys of the dict
                label: str
                original_image: ndarray
                subimages: list of dict
                    dict keys:
                        image: ndarray
                        mask: ndarray
                        position list of four size-two tuples
                        name: subimage name

        """

        # change the key name of the image data from the original cropped_image_data from image to original image
        for k in image_data_dict.keys():
            image_data_dict[k]['original_image'] = image_data_dict[k].pop('image')

        # preprocess the subimages
        image_data_dict = preprocess_subimages(image_data_dict, *args, **kwargs)

        # process the subimages if there are any
        print("z norming subimages")
        image_data_dict, self.subimage_mean, self.subimage_std = z_norm_subimages(image_data_dict, *args, **kwargs)

        print("transposing subimages")
        for k, x in image_data_dict.items():
            for s_image_name, s_image_data in image_data_dict[k]['sub_images'].items():
                image_data_dict[k]['sub_images'][s_image_name]['sub_image_cropped_padded_z_normed'] = s_image_data[
                    'sub_image_cropped_padded_z_normed'].transpose((2, 0, 1))

        # get rid of the extra fields
        print("rewriting dictionary keys")
        subimage_names = list(image_data_dict[list(image_data_dict.keys())[0]]['sub_images'].keys())
        for image_name, image_data in image_data_dict.items():
            subimages = image_data.pop('sub_images')
            image_data['sub_images'] = []
            for s_image_name in subimage_names:
                image_data['sub_images'].append(
                    {'image': subimages[s_image_name]['sub_image_cropped_padded_z_normed'],
                     'mask': subimages[s_image_name]['patch_mask'],
                     'position': subimages[s_image_name]['position'],
                     'name': s_image_name})
        self.image_data_dict = image_data_dict
        return image_data_dict

    def compute_perceptual_attention(self, image_name, source_attention=None, overlay_alpha=0.75, is_plot_results=True, save_dir=None,
                                     notes='', discard_ratio=0.9, *args, **kwargs):
        """

        Parameters
        ----------
        image_name: name of the image in the image data dict
        source_attention: default None, ndarray: the human attention with which the perceptual attention will be computed.
                        if not provided, the model attention will be returned
        is_plot_results: if True, the results will be plotted, see the parameter save_dir
        save_dir: if provided, the plots will be saved to this directory instead of being shown

        Returns
        -------
        {"original_image_attention": rollout_image, "subimage_attention": subimage_roll, "subimage_position": subimage_positions}
        """
        assert self.model is not None, "model must be provided by setting it to the model attribute of the SubimageHandler class"
        assert image_name in self.image_data_dict.keys(), f"image name {image_name} is not in the image data dict"
        sample = self.image_data_dict[image_name]
        if source_attention is not None:
            assert source_attention.shape == sample['original_image'].shape[:-1], f"source attention shape {source_attention.shape} does not match image shape {sample['original_image'].shape[:-1]}"
        image_original_size = sample['original_image'].shape[:-1]

        device = next(self.model.parameters()).device
        patch_size = self.model.patch_height, self.model.patch_width

        # run the model on the image
        image, *_ = collate_subimages([sample])
        image = any_image_to_tensor(image, device)
        subimage_masks = [x[0].detach().cpu().numpy() for x in image['masks']]  # the masks for the subimages in a a single image
        subimages = [x[0].detach().cpu().numpy() for x in image['subimages']]  # the subimages in a single image
        subimage_positions = [x['position'] for x in sample['sub_images']]

        if (image_name, discard_ratio) in self.attention_cache.keys():
            attention = self.attention_cache[(image_name, discard_ratio)]
        else:
            vit_rollout = VITAttentionRollout(self.model, device=device, attention_layer_name='attn_drop', discard_ratio=discard_ratio, *args, **kwargs)
            attention = vit_rollout(depth=self.model.depth, in_data=image, fixation_sequence=None, return_raw_attention=True)
            self.attention_cache[(image_name, discard_ratio)] = attention

        if source_attention is not None:
            attention = attention[1:, 1:]  # remove the first row and column of the attention, which is the class token
            source_attention_patchified = []
            for s_image in sample['sub_images']:
                s_source_attention = source_attention[
                                     s_image['position'][0][1]:(s_image['position'][0][1] + s_image['image'].shape[1]),
                                     s_image['position'][0][0]:(s_image['position'][0][0] + s_image['image'].shape[2])]
                # pad the source attention to the size of the subimage
                s_source_attention = np.pad(s_source_attention, ((0, s_image['image'].shape[1] - s_source_attention.shape[0]),
                                                                 (0, s_image['image'].shape[2] - s_source_attention.shape[1])),
                                            mode='constant', constant_values=0)
                # patchify the source attention
                s_source_attention_patches = patchify(s_source_attention, patch_size)
                s_source_attention_patches = s_source_attention_patches.reshape(s_source_attention_patches.shape[0], -1)
                s_source_attention_patches = np.mean(s_source_attention_patches, axis=1)
                source_attention_patchified.append(s_source_attention_patches)
            source_attention_patchified = np.concatenate(source_attention_patchified)
            # compute the perceptual attention

            # simple multiplication ##########################################
            # zero out the diagonal of the attention
            attention = attention * (1 - np.eye(len(attention)))
            attention = np.einsum('i,ij->j', source_attention_patchified, attention)

            # _attention = np.zeros(len(source_attention_patchified))
            # for i in range(len(source_attention_patchified)):
            #     _attention += source_attention_patchified[i] * attention[i, :]
            # attention = _attention

            # bayesian inverse relation ##########################################
            # attention = vit_rollout(depth=self.model.depth, in_data=image, fixation_sequence=None)
            # attention = attention / (source_attention_patchified + 1e-6)
            # attention = attention / np.max(attention)

            # sum across the columns ##########################################
            # attention = np.sum(attention, axis=0)
            # get the subimage attention from the source
        else:
            attention = attention[0, 1:]  # get the attention from the class token to the subimages
        rollout_image, subimage_roll = process_aoi(attention, image_original_size, True,
                                                   grid_size=self.model.get_grid_size(),
                                                   subimage_masks=subimage_masks, subimages=subimages,
                                                   subimage_positions=subimage_positions, patch_size=patch_size,
                                                   **kwargs)
        if is_plot_results:
            image_original = sample['original_image']
            image_original = cv2.cvtColor(image_original, cv2.COLOR_BGR2RGB)
            # cmap_name = register_cmap_with_alpha('viridis')
            plot_image_attention(image_original, rollout_image, source_attention, 'plasma',
                                 notes=f'{notes}{image_name}', overlay_alpha=overlay_alpha, save_dir=save_dir)
            plot_subimage_rolls(subimage_roll, subimages, subimage_positions, self.subimage_std, self.subimage_mean,
                                'plasma', notes=f"{notes}{image_name}", overlay_alpha=overlay_alpha, save_dir=save_dir)
        return {"original_image_attention": rollout_image, "subimage_attention": subimage_roll, "subimage_position": subimage_positions}



"""model.py - Model and module class for ViT.
   They are built to mirror those in the official Jax implementation.
"""

from typing import Optional
import torch
from torch import nn
from torch.nn import functional as F

from .transformer import Transformer
from .utils import load_pretrained_weights, as_tuple
from .configs import PRETRAINED_MODELS


class PositionalEmbedding1D(nn.Module):
    """Adds (optionally learned) positional embeddings to the inputs."""

    def __init__(self, seq_len, dim):
        super().__init__()
        self.pos_embedding = nn.Parameter(torch.zeros(1, seq_len, dim))
    
    def forward(self, x):
        """Input has shape `(batch_size, seq_len, emb_dim)`"""
        return x + self.pos_embedding


class ViT(nn.Module):
    """
    Args:
        name (str): Model name, e.g. 'B_16'
        pretrained (bool): Load pretrained weights
        in_channels (int): Number of channels in input data
        num_classes (int): Number of classes, default 1000

    References:
        [1] https://openreview.net/forum?id=YicbFdNTTy
    """

    def __init__(
        self, 
        name: Optional[str] = None, 
        pretrained: bool = False, 
        patches: int = 16,
        dim: int = 768,
        ff_dim: int = 3072,
        num_heads: int = 12,
        num_layers: int = 12,
        attention_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
        representation_size: Optional[int] = None,
        load_repr_layer: bool = False,
        classifier: str = 'token',
        positional_embedding: str = '1d',
        in_channels: int = 3, 
        image_size: Optional[int] = None,
        num_classes: Optional[int] = None,
        freeze=True
    ):
        super().__init__()

        # Configuration
        if name is None:
            check_msg = 'must specify name of pretrained model'
            assert not pretrained, check_msg
            # assert not resize_positional_embedding, check_msg
            assert not check_msg
            if num_classes is None:
                num_classes = 1000
            if image_size is None:
                image_size = 384
        else:  # load pretrained model
            assert name in PRETRAINED_MODELS.keys(), \
                'name should be in: ' + ', '.join(PRETRAINED_MODELS.keys())
            config = PRETRAINED_MODELS[name]['config']
            patches = config['patches']
            dim = config['dim']
            ff_dim = config['ff_dim']
            num_heads = config['num_heads']
            num_layers = config['num_layers']
            attention_dropout_rate = config['attention_dropout_rate']
            dropout_rate = config['dropout_rate']
            representation_size = config['representation_size']
            classifier = config['classifier']
            if image_size is None:
                image_size = PRETRAINED_MODELS[name]['image_size']
            if num_classes is None:
                num_classes = PRETRAINED_MODELS[name]['num_classes']
        self.image_size = image_size                

        # Image and patch sizes
        h, w = as_tuple(image_size)  # image sizes
        fh, fw = as_tuple(patches)  # patch sizes
        gh, gw = h // fh, w // fw  # number of patches
        seq_len = gh * gw

        # Patch embedding
        self.patch_embedding = nn.Conv2d(in_channels, dim, kernel_size=(fh, fw), stride=(fh, fw))

        # Class token
        if classifier == 'token':
            self.class_token = nn.Parameter(torch.zeros(1, 1, dim))
            seq_len += 1
        
        # Positional embedding
        if positional_embedding.lower() == '1d':
            self.positional_embedding = PositionalEmbedding1D(seq_len, dim)
        else:
            raise NotImplementedError()
        
        # Transformer
        self.transformer = Transformer(num_layers=num_layers, dim=dim, num_heads=num_heads, 
                                       ff_dim=ff_dim, dropout=dropout_rate)
        
        # Representation layer
        if representation_size and load_repr_layer:
            self.pre_logits = nn.Linear(dim, representation_size)
            pre_logits_size = representation_size
        else:
            pre_logits_size = dim

        # Classifier head
        self.norm = nn.LayerNorm(pre_logits_size, eps=1e-6)
        self.fc = nn.Linear(pre_logits_size, num_classes)

        # Initialize weights
        self.init_weights()
        
        # Load pretrained model
        if pretrained:
            pretrained_num_channels = 3
            pretrained_num_classes = PRETRAINED_MODELS[name]['num_classes']
            pretrained_image_size = PRETRAINED_MODELS[name]['image_size']
            load_pretrained_weights(
                self, name, 
                load_first_conv=(in_channels == pretrained_num_channels),
                load_fc=(num_classes == pretrained_num_classes),
                load_repr_layer=load_repr_layer,
                resize_positional_embedding=(image_size != pretrained_image_size),
            )
            if freeze:
                self.freeze_all_except_fc()
        
    @torch.no_grad()
    def init_weights(self):
        def _init(m):
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)  # _trunc_normal(m.weight, std=0.02)  # from .initialization import _trunc_normal
                if hasattr(m, 'bias') and m.bias is not None:
                    nn.init.normal_(m.bias, std=1e-6)  # nn.init.constant(m.bias, 0)
        self.apply(_init)
        nn.init.constant_(self.fc.weight, 0)
        nn.init.constant_(self.fc.bias, 0)
        nn.init.normal_(self.positional_embedding.pos_embedding, std=0.02)  # _trunc_normal(self.positional_embedding.pos_embedding, std=0.02)
        nn.init.constant_(self.class_token, 0)

    def forward(self, x):
        """Breaks image into patches, applies transformer, applies MLP head.

        Args:
            x (tensor): `b,c,fh,fw`
        """
        b, c, fh, fw = x.shape
        x = self.patch_embedding(x)  # b,d,gh,gw
        x = x.flatten(2).transpose(1, 2)  # b,gh*gw,d
        if hasattr(self, 'class_token'):
            x = torch.cat((self.class_token.expand(b, -1, -1), x), dim=1)  # b,gh*gw+1,d
        if hasattr(self, 'positional_embedding'): 
            x = self.positional_embedding(x)  # b,gh*gw+1,d 
        x = self.transformer(x)  # b,gh*gw+1,d
        if hasattr(self, 'pre_logits'):
            x = self.pre_logits(x)
            x = torch.tanh(x)
        if hasattr(self, 'fc'):
            x = self.norm(x)[:, 0]  # b,d
            x = self.fc(x)  # b,num_classes
        return x

    def freeze_all_except_fc(self):
        for p in self.patch_embedding.parameters():
            p.requires_grad = False
        for p in self.positional_embedding.parameters():
            p.requires_grad = False
        for p in self.transformer.parameters():
            p.requires_grad = False
        if 'pre_logits' in self.__dict__.keys():
            for p in self.pre_logits.parameters():
                p.requires_grad = False
        for p in self.norm.parameters():
            p.requires_grad = False
        self.class_token.requires_grad = False



import numpy as np
import torch
from torch import nn
from timm.models.vision_transformer import VisionTransformer, Block, Attention
from einops import rearrange, repeat
from einops.layers.torch import Rearrange



from eidl.Models.ExpertTimmVisionTransformer import ExpertTimmViTBlock


class ExpertTimmVisionTransformerSubimage(nn.Module):
    def __init__(self, vision_transformer: VisionTransformer, num_dim_fixation=2, num_lstm_layers=2, rnn_hidden_dim=64, fixation_conditioned=True):
        """
        Composite class extending the VisionTransformer from timm
        the lstm encodes the fixation sequence
        Parameters
        ----------
        vision_transformer
        rnn_hidden_dim
        fixation_conditioned: use the fixation sequence to condition the input
        """
        super().__init__()
        self.vision_transformer = vision_transformer
        self.depth = len(vision_transformer.blocks)
        self.num_classes = vision_transformer.num_classes  # copy the num classes from the VisionTransformer but not using its mlp

        self.lstm_hidden_dim = rnn_hidden_dim
        self.num_dim_fixation = num_dim_fixation
        self.num_lstm_layers = num_lstm_layers
        self.fixation_conditioned = fixation_conditioned

        # wrap the original blocks with our implementation to access the attention activation
        for i in range(len(self.vision_transformer.blocks)):
            self.vision_transformer.blocks[i] = ExpertTimmViTBlock(self.vision_transformer.blocks[i])

        # add the lstm layer to process the fixation sequence
        self.lstm = nn.LSTM(self.num_dim_fixation, self.lstm_hidden_dim, num_layers=self.num_lstm_layers, bidirectional=False, batch_first=True)

        if self.fixation_conditioned:
            self.vision_transformer.head = self.head = nn.Linear(
                self.vision_transformer.embed_dim + self.lstm_hidden_dim * self.num_lstm_layers,
                self.num_classes) if self.num_classes > 0 else nn.Identity()
        else:
            self.vision_transformer.head = self.head = nn.Linear(self.vision_transformer.embed_dim, self.num_classes) if self.num_classes > 0 else nn.Identity()

        # self.to_patch_embedding = nn.Sequential(
        #     Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1=self.vision_transformer.patch_embed.patch_size[0], p2=self.vision_transformer.patch_embed.patch_size[1]),
        #     nn.Linear(np.prod(self.vision_transformer.patch_embed.patch_size) * 3, self.vision_transformer.embed_dim),  # times three for color channels
        # )
        self.mask_token = nn.Parameter(torch.zeros(self.vision_transformer.embed_dim))


    def forward_features(self, img, collapse_attention_matrix=True, *args, **kwargs):
        subimage_xs = [self.vision_transformer.patch_embed(x) for x in img['subimages']]
        subimage_xs = [rearrange(x, 'b h w d -> b (h w) d') for x in subimage_xs]

        # concate the masks
        masks = torch.cat([rearrange(x, 'b h w -> b (h w)') for x in img['masks']], dim=1)
        x = torch.cat(subimage_xs, dim=1)
        # apply the mask token
        x[masks] = self.mask_token

        # recover a dummy w dimension for the _pos_embed in timm
        x = rearrange(x, 'b (h w) d -> b h w d', w=1)

        x = self.vision_transformer._pos_embed(x)  # TODO use a pos embedding in this class, not rely on dynamic image size in TIMM
        x = self.vision_transformer.norm_pre(x)
        # if self.grad_checkpointing and not torch.jit.is_scripting():
        #     x = checkpoint_seq(self.blocks, x)
        # else:
        x = self.vision_transformer.blocks(x)
        attention = self.vision_transformer.blocks[-1].attention  # we keep the attention activation of the last layer

        x = self.vision_transformer.norm(x)
        try:
            assert attention is not None
        except AssertionError:
            raise ValueError("the attention activation in forward_features is none, check your depth parameter")
        return x, attention

    def forward(self, img, fixation_sequence, collapse_attention_matrix=True, *args, **kwargs):
        x, attention = self.forward_features(img)

        if collapse_attention_matrix:
            attention = attention[:, :, 1:, 1:]  # take the self attention
            attention = attention / torch.sum(attention, dim=3, keepdim=True)
            attention = torch.sum(attention, dim=2)

        x = self.vision_transformer.forward_head(x, pre_logits=True)  # set pre_logits to true, don't apply the classification head
        if self.fixation_conditioned:
            rnn_outputs, hiddens = self.lstm(fixation_sequence)
            hidden, cell = hiddens
            x = torch.concat([x] + [hidden[i, :, :] for i in range(self.num_lstm_layers)], dim=1)
        x = self.head(x)
        return x, attention

    def get_grid_size(self):
        return self.vision_transformer.patch_embed.grid_size

    # TODO add test function, without fixation sequence as in ExpertAttentionViT.test()



import torch


def chunker(seq, size):
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


def reverse_tuple(t):
    if len(t) == 0:
        return t
    else:
        return(t[-1],)+reverse_tuple(t[:-1])


def collate_fn(batch):
    # label = torch.LongTensor([item['label'] for item in batch])
    label = torch.IntTensor([item['label_encoded'] for item in batch])
    label_encoded = torch.FloatTensor([item['label_onehot_encoded'] for item in batch])
    # if np.any(np.array([item['seq'] for item in batch]) == None):
    #     fixation_sequence = None
    #     aoi_heatmap = None
    # else:
    fixation_sequence = [torch.FloatTensor(item['fix_seq']) for item in batch]
    aoi_heatmap = torch.stack([torch.FloatTensor(item['aoi']) for item in batch], dim=0)
    image_resized = torch.stack([torch.FloatTensor(item['image']) for item in batch], dim=0)
    # image_original = torch.stack([torch.FloatTensor(item['original_image']) for item in batch], dim=0)
    image_original = [torch.FloatTensor(item['original_image']) for item in batch]

    if 'sub_images' in batch[0].keys():
        img, subimage_positions = collate_subimages(batch)
        return img, label, label_encoded, fixation_sequence, aoi_heatmap, image_resized, image_original, subimage_positions
    else:
        img = torch.stack([torch.FloatTensor(item['image_z_normed']) for item in batch], dim=0)
        return img, label, label_encoded, fixation_sequence, aoi_heatmap, image_resized, image_original


def collate_subimages(batch):
    subimages = []
    subimage_masks = []
    n_subimages = len(batch[0]['sub_images'])
    subimage_positions = []
    for i in range(n_subimages):
        subimages.append(torch.stack([torch.FloatTensor(item['sub_images'][i]['image']) for item in batch], dim=0))
        subimage_masks.append(torch.stack([torch.BoolTensor(item['sub_images'][i]['mask']) for item in batch], dim=0))
        subimage_positions.append([item['sub_images'][i]['position'] for item in batch])
    return {'subimages': subimages, 'masks': subimage_masks}, subimage_positions


import copy
import os.path
import pickle
import re
import time
from multiprocessing import Pool

import cv2
import numpy as np

import pandas as pd
import torch
import torch.nn.functional as F
from PIL import Image
from sklearn import preprocessing
from sklearn.model_selection import StratifiedShuffleSplit
from torch.utils.data import DataLoader
from torchvision import transforms, datasets


def load_image(image_path):
    return Image.open(image_path)

class ODIRDataset:
    """
    Ocular disease intelligent recognition dataset from: https://www.kaggle.com/datasets/andrewmvd/ocular-disease-recognition-odir5k
    Publication:
    A Benchmark of Ocular Disease Intelligent Recognition: One Shot for Multi-disease Detection
    https://link.springer.com/chapter/10.1007/978-3-030-71058-3_11
    """
    def __init__(self, dataset_root, image_size=None, n_jobs=16):
        """
        :param dataset_root: please set the dataset_root to the root of the extracted folder
        """
        # load the labels
        self.df = pd.read_csv(os.path.join(dataset_root, "full_df.csv"))

        # reload the entire dataset into memory for faster processing
        image_names = [(os.path.join(dataset_root, "preprocessed_images", row["filename"]), ) for index, row in self.df.iterrows()]
        self.labels = np.array([re.sub(r'\W+', '', row["labels"]) for index, row in self.df.iterrows()])
        print(f"Loading {len(image_names)} images with {n_jobs} jobs")
        start_time = time.time()
        with Pool(n_jobs) as p:
            self.img_data = p.starmap(load_image, image_names)
        print("Done loading images, took {} seconds".format(time.time() - start_time))
        # for index, row in self.df.iterrows():
        #     print("Loading {} of {} images".format(index + 1, len(self.df.index)), end='\r')
        #     img = Image.open(os.path.join(dataset_root, "preprocessed_images", row["filename"]))
        #     # img = np.moveaxis(img, -1, 0)
        #     label = row["labels"]
        #     self.img_data.append(img)
        #     self.labels.append(re.sub(r'\W+', '', label))
        # one-hot encode the labels
        self.num_class = len(np.unique(self.labels))
        # assert len(set([(x.shape[1]) for x in self.img_data] + [(x.shape[0]) for x in self.img_data])) == 1  # check image size
        self.image_size = image_size
        # self.img_data = np.array(self.img_data, dtype='uint8')  # set datatype to float32 to match that of torch's default datatype
        self.img_data_norm = None

        self.label_encoder = preprocessing.OneHotEncoder()
        self.onehot_encoded_labels = self.label_encoder.fit_transform(np.array(self.labels).reshape(-1, 1)).toarray()
        # print("Convert data to GPU tensor")
        # self._onehot_encoded_labels = torch.Tensor(self.onehot_encoded_labels).to(device)
        # self._img_data = torch.Tensor(self.img_data).to(device)

    def apply_subset(self, indices):
        rtn = copy.deepcopy(self)
        rtn.img_data = rtn.img_data[indices]
        rtn.img_data_norm = rtn.img_data_norm[indices]
        rtn.labels = rtn.labels[indices]
        rtn.onehot_encoded_labels = rtn.onehot_encoded_labels[indices]
        rtn.df = rtn.df.iloc[indices]
        return rtn

    def apply_transform(self, transform):
        self.img_data_norm = []
        print("Applying transform to images")
        for i in range(len(self.img_data)):
            self.img_data_norm.append(transform(self.img_data[i]))
            self.img_data[i] = np.array(self.img_data[i])
        self.img_data_norm = torch.stack(self.img_data_norm)

    def preprocess_all(self, normalize_mean=None, normalize_std=None, image_size=None):
        print("Normalizing input images")
        for i in range(len(self.img_data)):
            self.img_data[i] = np.array(self.img_data[i])
        self.img_data = np.array(self.img_data, dtype=np.float64)  # must NOT use float32, otherwise mean will be wronge due to overflow
        if normalize_std is None and normalize_mean is None:
            normalize_mean = np.mean(self.img_data.reshape(-1, self.img_data.shape[-1]), axis=0)
            normalize_std = np.std(self.img_data.reshape(-1, self.img_data.shape[-1]), axis=0)
        if image_size is not None:
            transform = transforms.Compose([transforms.Resize(image_size), transforms.ToTensor(), transforms.Normalize(normalize_mean, normalize_std)])
        else:
            transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize(normalize_mean, normalize_std)])
        self.img_data_norm = []
        for i in range(len(self.img_data)):
            self.img_data_norm.append(transform(self.img_data[i]))
        self.img_data_norm = torch.stack(self.img_data_norm).to(torch.float32)

        # self.img_data = self.img_data / 255.  # normalize the uint8 pixels to between 0 and 1
        # if normalize_std is None and normalize_mean is None:
        #     means = np.mean(self.img_data, axis=(0, 2, 3), keepdims=True)
        #     stds = np.std(self.img_data, axis=(0, 2, 3), keepdims=True)
        # else:
        #     means = np.reshape(normalize_mean, newshape=[1, 3, 1, 1])
        #     stds = np.reshape(normalize_std, newshape=[1, 3, 1, 1])
        # self.img_data_norm = ((self.img_data - means) / stds).astype(np.float32)

    def __len__(self):
        return len(self.img_data)

    def __getitem__(self, index):
        return {'image': self.img_data_norm[index], 'y': self.onehot_encoded_labels[index], 'image_original': self.img_data[index]}

def get_ODIRDataset(data_root, train_ratio, batch_size, data_export_root = None, normalize='data', normalize_mean=None, normalize_std=None, transform = None, image_size=None, random_seed=None):

    if data_export_root is not None and os.path.exists(dataset_path := os.path.join(data_export_root, 'ODIR_dataset.p')):
        print(f"Loading dataset from {dataset_path}")
        dataset = pickle.load(open(dataset_path, 'rb'))
    else:
        print(f"Creating dataset from {data_root}")
        dataset = ODIRDataset(data_root)
        if normalize == "data":
            dataset.preprocess_all()
        elif normalize == "provided":
            assert normalize_mean is not None and normalize_std is not None
            dataset.preprocess_all(normalize_mean=normalize_mean, normalize_std=normalize_std, image_size=image_size)
        elif normalize == "transform":
            assert transform is not None
            dataset.apply_transform(transform)
        else:
            raise Exception(f"Unrecognized normalize option {normalize}")
        if data_export_root is not None:
            dataset_path = os.path.join(data_export_root, 'ODIR_dataset.p')
            pickle.dump(dataset, open(dataset_path, 'wb'))

    train_size = int(train_ratio * len(dataset))
    val_size = len(dataset) - train_size
    # train_set, val_set = torch.utils.data.random_split(dataset, [train_size, val_size])

    skf = StratifiedShuffleSplit(n_splits=1, test_size=(1 - train_ratio), random_state=random_seed)
    train_val_image_indices, test_image_indices = [(train, test) for train, test in skf.split(dataset.img_data, dataset.labels)][0]  # split by image labels, not trials!

    train_set = dataset.apply_subset(train_val_image_indices)
    val_set = dataset.apply_subset(test_image_indices)

    train_data_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
    val_data_loader = DataLoader(val_set, batch_size=batch_size, shuffle=True)
    return train_data_loader, val_data_loader, train_size, val_size, dataset[0]['image'].shape, dataset.num_class


def get_test_dataset(data_root, batch_size):
    data_transforms = {
        'train': transforms.Compose([
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ]),
        'val': transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ]),
    }
    image_datasets = {x: datasets.ImageFolder(os.path.join(data_root, x),
                                              data_transforms[x])
                      for x in ['train', 'val']}
    dataloaders = {x: torch.utils.data.DataLoader(image_datasets[x], batch_size=batch_size,
                                                  shuffle=True, num_workers=4)
                   for x in ['train', 'val']}
    dataset_sizes = {x: len(image_datasets[x]) for x in ['train', 'val']}
    class_names = image_datasets['train'].classes

    input_shape = image_datasets['train'][0][0].shape

    return dataloaders['train'], dataloaders['val'], dataset_sizes['train'], dataset_sizes['val'], input_shape, len(class_names)

if __name__ == "__main__":
    data_root = "D:/Dropbox/Dropbox/ExpertViT/Datasets/ODIR"
    dataset = ODIRDataset(data_root, device='cpu')


import contextlib
import math
import os.path
import pickle

import numpy as np
import torch
from sklearn import metrics
from tqdm import tqdm
from torch import nn, autograd
import torch.nn.functional as F
import torch.nn.utils.rnn as rnn_utils
from eidl.utils.torch_utils import torch_wasserstein_loss, any_image_to_tensor
from eidl.viz.bad_gradient import is_bad_grad

def get_class_weight(labels, n_classes, smoothing_factor=0.1):
    """
    An example of one-hot encoded label array, the original labels are [0, 6]
    The corresponding cw is:
                 Count
    0 -> [1, 0]  100
    6 -> [0, 1]  200
    cw:  [3, 1.5]
    because pytorch treat [1, 0] as the first class and [0, 1] as the second class. However, the
    count for unique one-hot encoded label came out of np.unique is in the reverse order [0, 1] and [1, 0].
    the count needs to be reversed accordingly.

    TODO check when adding new classes
    @param convert_to_tensor:
    @param device:
    @return:
    """
    if len(labels.shape) == 2:  # if is onehot encoded
        unique_classes, counts = torch.unique(labels, return_counts=True, dim=0)
        class_frequencies = torch.zeros(n_classes).to(unique_classes.device)
        class_frequencies[torch.argmax(unique_classes, dim=1)] = counts.float()
        class_frequencies = torch.flip(class_frequencies, dims=[0])  # refer to docstring
    elif len(labels.shape) == 1:
        unique_classes, class_frequencies = torch.unique(labels, return_counts=True)
    else:
        raise ValueError("encoded labels should be either 1d or 2d array")
    if len(class_frequencies) == 1:  # when there is only one class in the dataset
        return None
    class_proportions = (class_frequencies + smoothing_factor) / len(labels)
    class_weights = 1 / class_proportions
    # class_weights[class_weights == torch.inf] = 0
    return class_weights  # reverse the class weights because

def run_one_epoch(mode, model: nn.Module, data_loader, device, n_classes, optimizer=None, criterion=nn.CrossEntropyLoss, pbar=None):
    if mode == 'train':
        model.train()
    else:
        model.eval()
    mini_batch_i = 0
    batch_losses = []
    num_correct_preds = 0
    y_all = None
    y_all_pred_postlogtis = None
    for batch_data in data_loader:
        x = batch_data['image']
        y = batch_data['y']

        if mode == 'train': optimizer.zero_grad()

        mini_batch_i += 1
        if pbar: pbar.update(1)

        y_pred = model(x.to(device))
        if isinstance(y_pred, tuple):
            y_pred, _ = y_pred
        y_pred_postlogits = F.softmax(y_pred, dim=1)
        # y_tensor = F.one_hot(y, num_classes=2).to(torch.float32).to(device)
        y_tensor = y.to(device).to(y_pred.dtype)

        class_weight = get_class_weight(y_tensor, n_classes=8)
        # loss = criterion(weight=class_weight)(y_tensor, y_pred)
        loss = criterion()(y_pred, y_tensor)

        if mode == 'train':
            loss.backward()
            optimizer.step()

        # add to y to compute auc
        y_all = np.concatenate([y_all, y.detach().cpu().numpy()]) if y_all is not None else y.detach().cpu().numpy()
        y_all_pred_postlogtis = np.concatenate([y_all_pred_postlogtis, y_pred_postlogits.detach().cpu().numpy()]) if y_all_pred_postlogtis is not None else y_pred_postlogits.detach().cpu().numpy()

        # measure accuracy
        num_correct_preds += torch.sum(torch.argmax(y_tensor, dim=1) == torch.argmax(y_pred, dim=1)).item()

        if pbar: pbar.set_description(f"{'Training' if mode == 'train' else 'Evaluating'} [{mini_batch_i}]: loss:{loss.item():.8f}")
        batch_losses.append(loss.item())
    acc = num_correct_preds / len(data_loader.dataset)
    auc = metrics.roc_auc_score(y_all, y_all_pred_postlogtis)
    return batch_losses, acc, auc


def train(model, optimizer: torch.optim.Optimizer, train_data_loader, val_data_loader, epochs, model_name, save_dir, n_classes, criterion=nn.CrossEntropyLoss):
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    train_losses = []
    train_accs = []
    val_losses = []
    val_accs = []
    val_aucs = []
    best_loss = np.inf
    training_histories = {}
    for epoch in range(epochs):
        pbar = tqdm(total=math.ceil(len(train_data_loader.dataset) / train_data_loader.batch_size),
                    desc='Training {}'.format(model_name))

        model.train()  # set the model in training model (dropout and batchnormal behaves differently in train vs. eval)
        train_batch_losses, train_acc, _ = run_one_epoch('train', model, train_data_loader, device, n_classes, optimizer, criterion, pbar)
        train_losses.append(np.mean(train_batch_losses))
        train_accs.append(train_acc)
        pbar.close()
        # scheduler.step()

        model.eval()
        with torch.no_grad():
            pbar = tqdm(total=math.ceil(len(val_data_loader.dataset) / val_data_loader.batch_size),  desc='Validating {}'.format(model_name))
            val_batch_losses, val_acc, val_auc = run_one_epoch('eval', model, val_data_loader, device, n_classes, optimizer,criterion, pbar)
            val_losses.append(np.mean(val_batch_losses))
            val_accs.append(val_acc)
            val_aucs.append(val_auc)
            pbar.close()
        lrl = [param_group['lr'] for param_group in optimizer.param_groups]
        lr = sum(lrl) / len(lrl)
        print(f"Epoch {epoch}: val auc = {val_aucs[-1]:.8f} train accuracy = {train_accs[-1]:.8f}, train loss={train_losses[-1]:.8f}; val accuracy = {val_accs[-1]:.8f}, val loss={val_losses[-1]:.8f}; LR={lr:.8f}")

        if val_losses[-1] < best_loss:
            torch.save(model.state_dict(), os.path.join(save_dir, model_name))
            print(
                'Best model loss improved from {} to {}, saved best model to {}'.format(best_loss, val_losses[-1],  save_dir))
            best_loss = val_losses[-1]

        # Save training histories after every epoch
        training_histories = {'loss_train': train_losses, 'acc_train': train_accs, 'loss_val': val_losses, 'acc_val': val_accs}
        pickle.dump(training_histories, open(os.path.join(save_dir, 'training_histories.pickle'), 'wb'))
    return training_histories

# def run_validation(model: nn.Module, val_loader, device, dist=None, alpha=None, model_config_string='', criterion=nn.CrossEntropyLoss):
#         model.eval()
#         total_samples = 0
#         total_loss = 0.0
#         total_correct = 0
#
#         pbar = tqdm(total=math.ceil(len(val_loader.dataset) / val_loader.batch_size),
#                     desc=f'Validating {model_config_string}')
#         pbar.update(mini_batch_i := 0)
#
#         for batch in val_loader:
#             mini_batch_i += 1
#             pbar.update(1)
#
#             image, label_encoded, label_onehot_encoded, fixation_sequence, aoi_heatmap, *_ = batch
#             fixation_sequence_torch = torch.Tensor(rnn_utils.pad_sequence(fixation_sequence, batch_first=True))
#             output, attention = model(image.to(device), fixation_sequence_torch.to(device))
#             # pred = F.softmax(output, dim=1)
#
#             aoi_heatmap = torch.flatten(aoi_heatmap, 1, 2)
#             attention = torch.sum(attention, dim=1)  # summation across the heads
#             attention /= torch.sum(attention, dim=1, keepdim=True)
#
#             y_tensor = label_onehot_encoded.to(device)
#             class_weight = get_class_weight(y_tensor, n_classes=2)
#             classification_loss = criterion(weight=class_weight)(output, y_tensor)
#
#             if dist is not None and alpha is not None:
#                 if dist == 'cross-entropy':
#                     attention_loss = alpha * F.cross_entropy(attention, aoi_heatmap.to(device))
#                 elif dist == 'Wasserstein':
#                     attention_loss = alpha * torch_wasserstein_loss(attention, aoi_heatmap.to(device))
#                 else:
#                     raise NotImplementedError(f" Loss type {dist} is not implemented")
#                 loss = classification_loss + attention_loss
#             else:
#                 loss = classification_loss
#
#             _, predictions = torch.max(F.softmax(output, dim=1), 1)
#             total_samples += (predictions.size(0))
#             total_loss += loss.item() * len(batch[0])
#             total_correct += torch.sum(predictions == label_encoded.to(device)).item()
#
#         epoch_loss = total_loss / total_samples
#         epoch_acc = (total_correct / total_samples)
#         pbar.close()
#
#         return epoch_loss, epoch_acc

def run_one_epoch_oct(mode, model: nn.Module, train_loader, optimizer, device, class_weights, model_config_string, criterion,
                      dist, alpha, l2_weight, epoch_i, *args, **kwargs):
    if mode == 'train':
        model.train()
    elif mode == 'val':
        model.eval()
    else:
        raise ValueError('mode must be train or eval')
    context_manager = torch.no_grad() if mode == 'val' else contextlib.nullcontext()

    total_samples = 0
    total_loss = 0.0
    total_correct = 0
    mini_batch_i = 0
    pbar = tqdm(total=math.ceil(len(train_loader.dataset) / train_loader.batch_size), desc=f'Training {model_config_string}')
    pbar.update(mini_batch_i)

    grad_norms = []  # debug
    for batch in train_loader:
        mini_batch_i += 1
        pbar.update(1)

        # prepare the input data ##############################################################
        image, label_encoded, label_onehot_encoded, fixation_sequence, aoi_heatmap, *_ = batch
        fixation_sequence_torch = torch.Tensor(rnn_utils.pad_sequence(fixation_sequence, batch_first=True))
        image = any_image_to_tensor(image, device)

        # check the aoi needs to be flattened
        if len(aoi_heatmap.shape) == 3:  # batch, height, width
            aoi_heatmap = torch.flatten(aoi_heatmap, 1, 2)

        # the forward pass ###################################################################
        if mode == 'train':
            optimizer.zero_grad()
        with context_manager:
            output, attention = model(image, fixation_sequence=fixation_sequence_torch.to(device))
            # pred = F.softmax(output, dim=1)

            attention = torch.sum(attention, dim=1)  # summation across the heads
            attention /= torch.sum(attention, dim=1, keepdim=True)  # normalize the attention output

            y_tensor = label_onehot_encoded.to(device)
            if class_weights is not None:
                classification_loss = criterion(weight=class_weights)(output, y_tensor)
            else:
                classification_loss = criterion()(output, y_tensor)

            if alpha is not None:
                if dist == 'cross-entropy':
                    attention_loss = alpha * F.cross_entropy(attention, aoi_heatmap.to(device))
                elif dist == 'Wasserstein':
                    attention_loss = alpha * torch_wasserstein_loss(attention, aoi_heatmap.to(device))
                else:
                    raise NotImplementedError(f" Loss type {dist} is not implemented")
            else:
                attention_loss = 0.

            if l2_weight:
                l2_penalty = l2_weight * sum([(p ** 2).sum() for p in model.parameters()])
                loss = classification_loss + attention_loss + l2_penalty
            else:
                loss = classification_loss + attention_loss

        # update the weights #################################################################
        if mode == 'train':
            loss.backward()

            # with autograd.detect_anomaly():
            #     try:
            #         loss.backward()
            #     except Exception as e:
            #         print(f"Bad gradient encountered: {e}")
            grad_norms.append([torch.mean(param.grad.norm()).item() for _, param in model.named_parameters() if param.grad is not None])

            # nn.utils.clip_grad_norm_(model.parameters(), max_norm=2.0, norm_type=2)
            nn.utils.clip_grad_value_(model.parameters(), clip_value=1.0)
            bad_grads = {}
            for name, param in model.named_parameters():
                if param.grad is not None and is_bad_grad(param.grad):
                    print(f"Find nan in param.grad in module: {name}")
                    bad_grads[name] = param.grad

            optimizer.step()
        ######################################################################################

        _, predictions = torch.max(F.softmax(output, dim=1), 1)
        total_samples += (predictions.size(0))
        total_loss += loss.item() * len(batch[0])
        total_correct += torch.sum(predictions == label_encoded.to(device)).item()
        pbar.set_description('Training Epoch-[{}]  Batch-[{}]: loss:{:.8f}, with classification loss {:.8f}, with attention loss {:.8f}'.format(epoch_i, mini_batch_i, loss.item(), classification_loss.item(), attention_loss.item()))

    epoch_loss = total_loss / total_samples
    epoch_acc = (total_correct / total_samples)
    pbar.close()
    return epoch_loss, epoch_acc

def train_oct_model(model, model_config_string, train_loader, valid_loader, results_dir,
                    criterion=nn.CrossEntropyLoss, num_epochs=100, alpha=0.01, l2_weight=None, dist='cross-entropy', *args, **kwargs):

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    best_acc = 0.0

    train_loss_list = []
    train_acc_list = []
    valid_loss_list = []
    valid_acc_list = []
    for epoch in range(num_epochs):
        print('epoch:{:d} / {:d}'.format(epoch, num_epochs))
        print('*' * 100)
        train_loss, train_acc = run_one_epoch_oct('train', model, train_loader, device=device, model_config_string=model_config_string, criterion=criterion,
                                                dist=dist, alpha=alpha, l2_weight=l2_weight, epoch_i=epoch, *args, **kwargs)
        train_loss_list.append(train_loss)
        train_acc_list.append(train_acc)
        valid_loss, valid_acc = run_one_epoch_oct('val', model, valid_loader, device=device, model_config_string=model_config_string, criterion=criterion,
                                                  dist=dist, alpha=alpha, l2_weight=l2_weight, epoch_i=epoch,*args, **kwargs)
        valid_loss_list.append(valid_loss)
        valid_acc_list.append(valid_acc)
        print("training loss: {:.4f}, training acc: {:.4f}; validation loss {:.4f}, validation acc: {:.4f}".format(train_loss, train_acc, valid_loss, valid_acc))

        with open(os.path.join(results_dir, f'log_{model_config_string}.txt'), 'a+') as file:
            file.write('epoch:{:d} / {:d}\n'.format(epoch, num_epochs))
            file.write("training: {:.4f}, {:.4f}\n".format(train_loss, train_acc))
            file.write("validation: {:.4f}, {:.4f}\n".format(valid_loss, valid_acc))
        file.close()

        if valid_acc > best_acc:
            best_acc = valid_acc
            best_model = model
            torch.save(best_model, os.path.join(results_dir, f'best_{model_config_string}.pt'))
            torch.save(model.state_dict(), os.path.join(results_dir, f'best_{model_config_string}_statedict.pt'))

        # if epoch >= 10 and len(set(train_acc_list[-10:])) == 1 and len(set(valid_acc_list[-10:])) == 1:
        #     break
    torch.save(model, os.path.join(results_dir, f'final_{model_config_string}.pt'))
    return train_loss_list, train_acc_list, valid_loss_list, valid_acc_list

# def test_without_fixation(model, data_loader, device):
#     total_samples = 0
#     total_correct = 0
#     model.train(False)
#     predicted_attentions = []
#     pbar = tqdm(total=math.ceil(len(data_loader.dataset) / data_loader.batch_size),
#                 desc=f'Testing without fixaiton')
#     pbar.update(mini_batch_i := 0)
#
#     with torch.no_grad():
#         for batch in data_loader:
#             mini_batch_i += 1
#             pbar.update(1)
#
#             image, label, label_encoded, fixation_sequence, aoi_heatmap = batch
#             output, attention = model.test(image.to(device))
#             pred = F.softmax(output, dim=1)
#
#             predicted_attentions.append(attention.detach().cpu().numpy())
#             _, predictions = torch.max(pred, 1)
#             total_samples += (predictions.size(0))
#             total_correct += torch.sum(predictions == label.to(device)).item()
#
#     test_acc = (total_correct / total_samples)
#     return test_acc


import os.path

import cv2
import numpy as np
from matplotlib import pyplot as plt


def plt2arr(fig, draw=True):
    """
    need to draw if figure is not drawn yet
    """
    if draw:
        fig.canvas.draw()
    rgba_buf = fig.canvas.buffer_rgba()
    (w,h) = fig.canvas.get_width_height()
    rgba_arr = np.frombuffer(rgba_buf, dtype=np.uint8).reshape((h,w,4))
    return rgba_arr

def plot_train_history(history, note=''):
    plt.plot(history['train_accs'])
    plt.plot(history['val_accs'])
    plt.title('model accuracy ' + note)
    plt.ylabel('accuracy')
    plt.xlabel('epoch')
    plt.legend(['train', 'test'], loc='lower left')
    plt.show()

    # summarize history for loss
    plt.plot(history['train_losses'])
    plt.plot(history['val_losses'])
    plt.title('model loss ' + note)
    plt.ylabel('loss')
    plt.xlabel('epoch')
    plt.legend(['train', 'test'], loc='lower left')
    plt.show()


def plot_subimage_rolls(subimage_roll, subimages, subimage_positions, image_std, image_mean, cmap_name,
                        notes='', overlay_alpha=0.75, save_dir=None):
    for s_i, (s_roll, s_image, s_pos) in enumerate(zip(subimage_roll, subimages, subimage_positions)):
        # unznorm the image
        s_image_unznormed = np.transpose(s_image, (1, 2, 0)) * image_std + image_mean
        s_image_unznormed = s_image_unznormed.astype(np.uint8)
        s_image_unznormed = cv2.cvtColor(s_image_unznormed, cv2.COLOR_BGR2RGB)
        s_image_size = s_pos[2][1] - s_pos[0][1], s_pos[2][0] - s_pos[0][0]
        s_image_unznormed = s_image_unznormed[:s_image_size[0], :s_image_size[1]]

        # plot the aoi and subimage side by side, using subplot
        s_fig = plt.figure(figsize=(15, 10), constrained_layout=True)

        plt.subplot(1, 3, 1)
        plt.imshow(s_image_unznormed, cmap=cmap_name)
        plt.imshow(s_roll, cmap=cmap_name, alpha=overlay_alpha * s_roll / np.max(s_roll))
        plt.axis('off')

        plt.subplot(1, 3, 2)
        plt.imshow(s_image_unznormed)
        plt.axis('off')

        plt.subplot(1, 3, 3)
        plt.imshow(s_roll, cmap=cmap_name)
        plt.colorbar()
        plt.axis('off')

        plt.suptitle(title_text := f'{notes} Subimage {s_i}')

        if save_dir is not None:
            plt.savefig(os.path.join(save_dir, f'{title_text}.png'))
        else:
            plt.show()

def plot_image_attention(image_original, model_attention, source_attention, cmap_name, overlay_alpha=0.9, save_dir=None, notes=''):
    fig = plt.figure(figsize=(30, 20), constrained_layout=True)

    plt.subplot(2, 2, 1)
    plt.imshow(image_original)  # plot the original image
    if source_attention is not None:
        plt.imshow(source_attention, cmap=cmap_name, alpha=overlay_alpha * source_attention / np.max(source_attention))
    plt.axis('off')
    plt.title("Source Attention Overlay")

    if source_attention is not None:
        plt.subplot(2, 2, 3)
        plt.imshow(source_attention, cmap=cmap_name)
        plt.axis('off')
        plt.title("Source Attention")

    plt.subplot(2, 2, 2)
    plt.imshow(image_original)  # plot the original image
    plt.imshow(model_attention, cmap=cmap_name, alpha=overlay_alpha * model_attention / np.max(model_attention))
    plt.axis('off')
    plt.title("Model Attention Overlay")

    plt.subplot(2, 2, 4)
    plt.imshow(model_attention, cmap=cmap_name)
    plt.axis('off')
    plt.title("Model Attention")

    plt.suptitle(notes)
    # plt.show()

    if save_dir is not None:
        fig.savefig(os.path.join(save_dir, f'{notes}.png'))
    else:
        plt.show()



import os
import pickle

import cv2
import imageio
import matplotlib
import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from torch.utils.data import DataLoader
import torch.nn.utils.rnn as rnn_utils
from PIL import Image

from eidl.utils.image_utils import remap_subimage_aoi, process_aoi
from eidl.utils.iter_utils import chunker, collate_fn
from eidl.utils.model_utils import parse_model_parameter, get_best_model, parse_training_results
from eidl.utils.torch_utils import any_image_to_tensor
from eidl.viz.vit_rollout import VITAttentionRollout

from eidl.viz.viz_utils import plt2arr, plot_train_history, plot_subimage_rolls, plot_image_attention


def register_cmap_with_alpha(cmap_name):
    # get colormap
    ncolors = 256
    color_array = plt.get_cmap(cmap_name)(range(ncolors))
    # change alpha values
    color_array[:, -1] = np.linspace(1.0, 0.0, ncolors)
    # create a colormap object
    cmap_rtn = f'{cmap_name}_alpha'
    map_object = LinearSegmentedColormap.from_list(name=cmap_rtn, colors=color_array)
    # register this new colormap with matplotlib
    plt.register_cmap(cmap=map_object)
    return cmap_rtn

def viz_oct_results(results_dir, batch_size, n_jobs=1, acc_min=.3, acc_max=1, viz_val_acc=True, plot_format='individual', num_plot=14, rollout_transparency=0.75):
    '''

    Parameters
    ----------
    results_dir
    test_image_path
    test_image_main
    batch_size
    image_size
    n_jobs
    acc_min
    acc_max
    viz_val_acc
    plot_format: can be 'individual' or 'grid'. Note setting to 'grid' will not plot the gifs
    num_plot

    Returns
    -------

    '''

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda:0" if use_cuda else "cpu")

    image_stats = pickle.load(open(os.path.join(results_dir, 'image_stats.p'), 'rb'))
    # load the test dataset ############################################################################################
    test_dataset = pickle.load(open(os.path.join(results_dir, 'test_dataset.p'), 'rb'))

    results_dict, model_config_strings = parse_training_results(results_dir)

    # results_df.to_csv(os.path.join(results_dir, "summary.csv"))

    # visualize the val acc across alpha ###############################################################################
    alphas = {parse_model_parameter(x, 'alpha') for x in model_config_strings}
    alphas = list(alphas)
    alphas.sort()
    models = {parse_model_parameter(x, 'model') for x in model_config_strings}

    small_font_size = 24
    medium_font_size = 26
    large_font_size = 30

    plt.rc('font', size=small_font_size)
    plt.rc('axes', titlesize=small_font_size)
    plt.rc('axes', labelsize=small_font_size)
    plt.rc('xtick', labelsize=small_font_size)
    plt.rc('ytick', labelsize=small_font_size)
    plt.rc('legend', fontsize=small_font_size)
    plt.rc('figure', titlesize=large_font_size)

    if viz_val_acc:
        fig = plt.figure(figsize=(15, 10), constrained_layout=True)
        xticks = np.array(list(range(1, len(alphas) + 1)))
        model_x_offset = 0.3
        box_width = 0.25
        colors = matplotlib.cm.tab20(range(20))

        for i, model in enumerate(models):
            val_accs = []
            for alpha in alphas:
                val_acc_alpha = []
                for model_config_string, results in results_dict.items():
                    if parse_model_parameter(model_config_string, 'alpha') == alpha and parse_model_parameter(model_config_string, 'model') == model:
                        val_acc_alpha.append(np.max(results['val_accs']))
                val_accs.append(val_acc_alpha)
            x_positions = xticks + model_x_offset * i
            plt.boxplot(val_accs, positions=x_positions, patch_artist=True, widths=box_width, boxprops=dict(facecolor=colors[i*2+1], alpha=0.5, color=colors[i*2]), whiskerprops=dict(color=colors[i*2]), capprops=dict(color=colors[i*2]), medianprops=dict(color=colors[i*2]))
            plt.plot(x_positions, [np.mean(x) for x in val_accs], label=f"{model} average across tested parameters", color=colors[i*2])
            plt.scatter(x_positions, [np.mean(x) for x in val_accs], color=colors[i*2], s=40)

        plt.ylim(acc_min, acc_max)
        plt.xticks(ticks=xticks, labels=alphas)
        plt.xlabel("Expert AOI weight (α)")
        plt.ylabel("Validation accuracy")
        plt.title(f"validation accuracy across expert AOI weights")
        plt.legend()
        plt.tight_layout()
        plt.show()

        # visualize the hyperparam space ##################################################################################
        parameter_to_test_base = 'lr', 'depth', 'dist'
        parameter_to_test_pretrained = 'lr', 'dist'
        xticks = np.array(list(range(len(alphas))))
        for i, model in enumerate(models):
            for hyperparam_name in parameter_to_test_base if model == 'base' else parameter_to_test_pretrained:
                hyperparam_space = {parse_model_parameter(x, hyperparam_name) for x in model_config_strings if model == parse_model_parameter(x, 'model')}
                hyperparam_space = list(hyperparam_space)
                if isinstance(hyperparam_space[0], float):
                    hyperparam_space.sort()
                fig = plt.figure(figsize=(15, 10), constrained_layout=True)
                val_accs = np.empty((len(hyperparam_space), len(alphas)))
                for j, hyper_param in enumerate(hyperparam_space):
                    for k, alpha in enumerate(alphas):
                        val_acc_hyperparam_alpha = []
                        for model_config_string, results in results_dict.items():
                            if parse_model_parameter(model_config_string, hyperparam_name) == hyper_param and parse_model_parameter(model_config_string, 'alpha') == alpha and parse_model_parameter(model_config_string, 'model') == model:
                                val_acc_hyperparam_alpha.append(np.max(results['val_accs']))
                        val_accs[j, k] = np.mean(val_acc_hyperparam_alpha)
                        plt.text(k, j, round(float(np.mean(val_acc_hyperparam_alpha)), 3))
                plt.imshow(val_accs, vmin=acc_min, vmax=acc_max)
                plt.xticks(ticks=xticks, labels=alphas)
                plt.yticks(ticks=list(range(len(hyperparam_space))), labels=[float('%.1g' % x) if isinstance(x, float) else x for x in hyperparam_space ])  # additional float casting to avoid e notation
                plt.xlabel("Expert AOI weight (α)")
                plt.ylabel(hyperparam_name)
                plt.colorbar()
                plt.title(f"{model}: validation accuracy for {hyperparam_name}-alpha ")
                plt.show()

    # visualize the attention rollout ##################################################################################
    cmap_name = register_cmap_with_alpha('viridis')

    models = list(reversed(list(models)))

    best_model, best_model_results, best_model_config_string = get_best_model(models, results_dict)
    best_model.eval()
    patch_size = best_model.patch_height, best_model.patch_width
    model_depth = best_model.depth

    # visualize the training history of the best model ##################################################################
    plot_train_history(best_model_results, note=f"{best_model_config_string}")

    # register target cmap #########################################################
    has_subimage = test_dataset.trial_samples[0].keys()
    test_loader = DataLoader(test_dataset, batch_size=1, shuffle=False, collate_fn=collate_fn)  # one image at a time

    with torch.no_grad():
        test_dataset.create_aoi(best_model.get_grid_size())
        # epoch_loss, epoch_acc = run_validation(best_model, test_loader, device)
        # print(f"Test acc: {epoch_acc}")

        # use gradcam is model is not a ViT
        vit_rollout = VITAttentionRollout(best_model, device=device, attention_layer_name='attn_drop', head_fusion="mean", discard_ratio=0.1)
        sample_count = 0

        if plot_format == 'grid':
            fig, axs = plt.subplots(model_depth + 2, num_plot, figsize=(2 * num_plot, 2 * (model_depth + 2)))
            plt.setp(axs, xticks=[], yticks=[])
            plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.05, hspace=0.05)
            fig.tight_layout()

        for batch in test_loader:
            print(f'Processing sample {sample_count} in test set')
            image, label, label_encoded, fix_sequence, aoi_heatmap, image_resized, image_original, *rest = batch
            if has_subimage:
                # take out the batches
                subimage_positions = [x[0] for x in rest[0]]
                subimage_masks = [x[0].detach().cpu().numpy() for x in image['masks']]  # the masks for the subimages in a a single image
                subimages = [x[0].detach().cpu().numpy() for x in image['subimages']]  # the subimages in a single image
            else:
                subimage_masks = None
                subimage_positions = None

            rolls = []
            fixation_sequence_torch = torch.Tensor(rnn_utils.pad_sequence(fix_sequence, batch_first=True))

            for roll_depth in range(best_model.depth):
                image = any_image_to_tensor(image, device)
                rolls.append(vit_rollout(depth=roll_depth, in_data=image, fixation_sequence=fixation_sequence_torch))

            image_original = np.array(image_original[0].numpy(), dtype=np.uint8)
            image_original = cv2.cvtColor(image_original, cv2.COLOR_BGR2RGB)
            image_original_size = image_original.shape[:2]
            if plot_format == 'individual':
                fig_list = []
                # plot the original image
                fig = plt.figure(figsize=(15, 10), constrained_layout=True)
                plt.imshow(image_original)  # plot the original image, bgr to rgb
                plt.axis('off')
                plt.title(f'#{sample_count}, original image')
                # plt.show()
                Image.fromarray(image_original).save(f'figures/Sample {sample_count} in test set, original image.png')
                fig_list.append(plt2arr(fig))
                # plot the original image with expert AOI heatmap
                fig = plt.figure(figsize=(15, 10), constrained_layout=True)
                plt.imshow(image_original)  # plot the original image

                _aoi_heatmap, *_ = process_aoi(aoi_heatmap[0].numpy(), image_original_size, has_subimage, grid_size=best_model.get_grid_size(),
                                           subimage_masks=subimage_masks, subimages=subimages, subimage_positions=subimage_positions, patch_size=patch_size)
                plt.imshow(_aoi_heatmap, cmap=cmap_name, alpha=rollout_transparency)
                plt.axis('off')
                plt.title(f'#{sample_count}, expert AOI')
                # plt.show()
                fig.savefig(f'figures/Sample {sample_count} in test set, expert attention.png')

                for i, roll in enumerate(rolls):
                    rollout_image, subimage_roll = process_aoi(rolls[0], image_original_size, has_subimage,
                                               grid_size=best_model.get_grid_size(),
                                               subimage_masks=subimage_masks, subimages=subimages, subimage_positions=subimage_positions, patch_size=patch_size)

                    plot_image_attention(image_original, rollout_image, _aoi_heatmap, cmap_name,
                                         notes=f'Sample  {sample_count} in test  set, model {model}, roll depth {i}')
                    plot_subimage_rolls(subimage_roll, subimages, subimage_positions, image_stats['std'], image_stats['mean'], cmap_name,
                                        notes=f"Sample {sample_count} in test set, model: {model}, roll depth {i}", overlay_alpha=rollout_transparency, save_dir='figures')

                    # fig.savefig(f'figures/valImageIndex-{sample_count}_model-{model}_rollDepth-{i}.png')
                    fig_list.append(plt2arr(fig))
                # imageio.mimsave(f'gifs/model-{model}_valImageIndex-{sample_count}.gif', fig_list, fps=2)  # TODO expose save dir
            elif plot_format == 'grid' and sample_count < num_plot:
                    axis_original_image, axis_aoi_heatmap, axes_roll = axs[0, sample_count], axs[1, sample_count], axs[2:, sample_count]
                    axis_original_image.imshow(image_original)  # plot the original image
                    axis_original_image.axis('off')
                    # axis_original_image.title(f'#{sample_count}, original image')

                    # plot the original image with expert AOI heatmap
                    axis_aoi_heatmap.imshow(image_original)  # plot the original image
                    _aoi_heatmap = cv2.resize(aoi_heatmap.numpy(), dsize=image.shape[1:], interpolation=cv2.INTER_LANCZOS4)
                    axis_aoi_heatmap.imshow(_aoi_heatmap.T, cmap=cmap_name, alpha=rollout_transparency)
                    axis_aoi_heatmap.axis('off')
                    # axis_aoi_heatmap.title(f'#{sample_count}, expert AOI')

                    for i, roll in enumerate(rolls):
                        rollout_image = cv2.resize(roll, dsize=image.shape[1:], interpolation=cv2.INTER_LANCZOS4)
                        axes_roll[i].imshow(np.moveaxis(image_resized, 0, 2))  # plot the original image
                        axes_roll[i].imshow(rollout_image.T, cmap=cmap_name, alpha=rollout_transparency)
                        axes_roll[i].axis('off')
                        # axes_roll[i].title(f'#{sample_count}, model {model}, , roll depth {i}')
            sample_count += 1

    if plot_format == 'grid':
        plt.show()






from eidl.viz.viz_oct_results import viz_oct_results

results_dir = '../temp/results-01_07_2024_10_53_56'

batch_size = 8

viz_val_acc = True

if __name__ == '__main__':
    viz_oct_results(results_dir, batch_size, viz_val_acc=viz_val_acc, plot_format='individual')


import torch
from torch import nn
from timm.models.vision_transformer import VisionTransformer, Block, Attention

class ExpertTimmVisionTransformer(nn.Module):
    def __init__(self, vision_transformer: VisionTransformer, num_dim_fixation=2, num_lstm_layers=2, rnn_hidden_dim=64, fixation_conditioned=True):
        """
        Composite class extending the VisionTransformer from timm
        the lstm encodes the fixation sequence
        Parameters
        ----------
        vision_transformer
        rnn_hidden_dim
        fixation_conditioned: use the fixation sequence to condition the input
        """
        super().__init__()
        self.vision_transformer = vision_transformer
        self.depth = len(vision_transformer.blocks)
        self.num_classes = vision_transformer.num_classes  # copy the num classes from the VisionTransformer but not using its mlp

        self.lstm_hidden_dim = rnn_hidden_dim
        self.num_dim_fixation = num_dim_fixation
        self.num_lstm_layers = num_lstm_layers
        self.fixation_conditioned = fixation_conditioned

        # wrap the original blocks with our implementation to access the attention activation
        for i in range(len(self.vision_transformer.blocks)):
            self.vision_transformer.blocks[i] = ExpertTimmViTBlock(self.vision_transformer.blocks[i])

        # add the lstm layer to process the fixation sequence
        self.lstm = nn.LSTM(self.num_dim_fixation, self.lstm_hidden_dim, num_layers=self.num_lstm_layers, bidirectional=False, batch_first=True)

        if self.fixation_conditioned:
            self.vision_transformer.head = self.head = nn.Linear(
                self.vision_transformer.embed_dim + self.lstm_hidden_dim * self.num_lstm_layers,
                self.num_classes) if self.num_classes > 0 else nn.Identity()
        else:
            self.vision_transformer.head = self.head = nn.Linear(self.vision_transformer.embed_dim, self.num_classes) if self.num_classes > 0 else nn.Identity()

    def forward_features(self, img, collapse_attention_matrix=True, *args, **kwargs):
        x = self.vision_transformer.patch_embed(img)
        x = self.vision_transformer._pos_embed(x)
        x = self.vision_transformer.norm_pre(x)
        # if self.grad_checkpointing and not torch.jit.is_scripting():
        #     x = checkpoint_seq(self.blocks, x)
        # else:
        x = self.vision_transformer.blocks(x)
        attention = self.vision_transformer.blocks[-1].attention  # we keep the attention activation of the last layer

        x = self.vision_transformer.norm(x)
        try:
            assert attention is not None
        except AssertionError:
            raise ValueError("the attention activation in forward_features is none, check your depth parameter")
        return x, attention

    def forward(self, img, fixation_sequence):
        x, attention = self.forward_features(img)
        attention = attention[:, :, 1:, 1:]
        attention = attention / torch.sum(attention, dim=3, keepdim=True)
        attention = torch.sum(attention, dim=2)

        x = self.vision_transformer.forward_head(x, pre_logits=True)  # set pre_logits to true, don't apply the classification head
        if self.fixation_conditioned:
            rnn_outputs, hiddens = self.lstm(fixation_sequence)
            hidden, cell = hiddens
            x = torch.concat([x] + [hidden[i, :, :] for i in range(self.num_lstm_layers)], dim=1)
        x = self.head(x)
        return x, attention

    def get_grid_size(self):
        return self.vision_transformer.patch_embed.grid_size

    # TODO add test function, without fixation sequence as in ExpertAttentionViT.test()

class ExpertTimmViTBlock(nn.Module):
    def __init__(self, block: Block):
        super().__init__()
        self.block = block
        self.attn = ExpertTimmViTAttention(self.block.attn)
        self.attention = None

    def forward(self, x):
        x, self.attention = self.attn(x)
        x = x + self.block.drop_path1(self.block.norm1(x))
        x = x + self.block.drop_path2(self.block.norm2(self.block.mlp(x)))
        return x

class ExpertTimmViTAttention(nn.Module):
    def __init__(self, attention: Attention):
        super().__init__()
        self.attention = attention

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.attention.qkv(x).reshape(B, N, 3, self.attention.num_heads, C // self.attention.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.attention.scale
        attn = attn.softmax(dim=-1)
        attn = self.attention.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.attention.proj(x)
        x = self.attention.proj_drop(x)
        return x, attn
